diff --git a/IMAGES/BH Graph.png b/IMAGES/BH Graph.png
new file mode 100644
index 000000000..03a69e512
Binary files /dev/null and b/IMAGES/BH Graph.png differ
diff --git a/IMAGES/BVH Graph.png b/IMAGES/BVH Graph.png
new file mode 100644
index 000000000..2bd9695cc
Binary files /dev/null and b/IMAGES/BVH Graph.png differ
diff --git a/IMAGES/BlackHoleSolo_V1.png b/IMAGES/BlackHoleSolo_V1.png
new file mode 100644
index 000000000..89c41e9de
Binary files /dev/null and b/IMAGES/BlackHoleSolo_V1.png differ
diff --git a/IMAGES/LightWarpSolo_V1.png b/IMAGES/LightWarpSolo_V1.png
new file mode 100644
index 000000000..9880e272c
Binary files /dev/null and b/IMAGES/LightWarpSolo_V1.png differ
diff --git a/IMAGES/StreamCompactGraph_V1.png b/IMAGES/StreamCompactGraph_V1.png
new file mode 100644
index 000000000..f49e41235
Binary files /dev/null and b/IMAGES/StreamCompactGraph_V1.png differ
diff --git a/IMAGES/Thin_Lens.png b/IMAGES/Thin_Lens.png
new file mode 100644
index 000000000..5c1a4dd61
Binary files /dev/null and b/IMAGES/Thin_Lens.png differ
diff --git a/IMAGES/blackhole_bvh_test.2025-10-05_02-25-49z.537samp.png b/IMAGES/blackhole_bvh_test.2025-10-05_02-25-49z.537samp.png
new file mode 100644
index 000000000..d5bd5ba71
Binary files /dev/null and b/IMAGES/blackhole_bvh_test.2025-10-05_02-25-49z.537samp.png differ
diff --git a/IMAGES/blackhole_bvh_test.2025-10-05_02-39-31z.775samp.png b/IMAGES/blackhole_bvh_test.2025-10-05_02-39-31z.775samp.png
new file mode 100644
index 000000000..505ed6d1b
Binary files /dev/null and b/IMAGES/blackhole_bvh_test.2025-10-05_02-39-31z.775samp.png differ
diff --git a/IMAGES/blackhole_mirrors.2025-10-05_13-27-25z.678samp.png b/IMAGES/blackhole_mirrors.2025-10-05_13-27-25z.678samp.png
new file mode 100644
index 000000000..77b6da27d
Binary files /dev/null and b/IMAGES/blackhole_mirrors.2025-10-05_13-27-25z.678samp.png differ
diff --git a/IMAGES/blackhole_mirrors.2025-10-08_03-18-09z.2523samp.png b/IMAGES/blackhole_mirrors.2025-10-08_03-18-09z.2523samp.png
new file mode 100644
index 000000000..4404b147a
Binary files /dev/null and b/IMAGES/blackhole_mirrors.2025-10-08_03-18-09z.2523samp.png differ
diff --git a/IMAGES/blackhole_mirrors.2025-10-08_03-24-26z.3236samp.png b/IMAGES/blackhole_mirrors.2025-10-08_03-24-26z.3236samp.png
new file mode 100644
index 000000000..c2a38cdee
Binary files /dev/null and b/IMAGES/blackhole_mirrors.2025-10-08_03-24-26z.3236samp.png differ
diff --git a/IMAGES/bvh_stress_test.2025-10-08_01-15-18z.1560samp.png b/IMAGES/bvh_stress_test.2025-10-08_01-15-18z.1560samp.png
new file mode 100644
index 000000000..dec79b664
Binary files /dev/null and b/IMAGES/bvh_stress_test.2025-10-08_01-15-18z.1560samp.png differ
diff --git a/IMAGES/bvh_stress_test.2025-10-08_01-25-28z.580samp.png b/IMAGES/bvh_stress_test.2025-10-08_01-25-28z.580samp.png
new file mode 100644
index 000000000..bd653471b
Binary files /dev/null and b/IMAGES/bvh_stress_test.2025-10-08_01-25-28z.580samp.png differ
diff --git a/IMAGES/compaction_test.2025-09-23_14-55-01z.5000samp.png b/IMAGES/compaction_test.2025-09-23_14-55-01z.5000samp.png
new file mode 100644
index 000000000..32e6f00d3
Binary files /dev/null and b/IMAGES/compaction_test.2025-09-23_14-55-01z.5000samp.png differ
diff --git a/IMAGES/compaction_test.2025-09-26_14-23-01z.3240samp.png b/IMAGES/compaction_test.2025-09-26_14-23-01z.3240samp.png
new file mode 100644
index 000000000..67c21cd41
Binary files /dev/null and b/IMAGES/compaction_test.2025-09-26_14-23-01z.3240samp.png differ
diff --git a/IMAGES/cornell.2025-09-24_00-09-51z.247samp.png b/IMAGES/cornell.2025-09-24_00-09-51z.247samp.png
new file mode 100644
index 000000000..4a08e5917
Binary files /dev/null and b/IMAGES/cornell.2025-09-24_00-09-51z.247samp.png differ
diff --git a/IMAGES/cornell.2025-10-03_20-11-58z.634samp.png b/IMAGES/cornell.2025-10-03_20-11-58z.634samp.png
new file mode 100644
index 000000000..cd2b3c5f7
Binary files /dev/null and b/IMAGES/cornell.2025-10-03_20-11-58z.634samp.png differ
diff --git a/IMAGES/cornell.2025-10-03_20-22-33z.363samp.png b/IMAGES/cornell.2025-10-03_20-22-33z.363samp.png
new file mode 100644
index 000000000..4de47c44f
Binary files /dev/null and b/IMAGES/cornell.2025-10-03_20-22-33z.363samp.png differ
diff --git a/IMAGES/cornell.2025-10-04_00-38-55z.35samp.png b/IMAGES/cornell.2025-10-04_00-38-55z.35samp.png
new file mode 100644
index 000000000..4c6f0e44e
Binary files /dev/null and b/IMAGES/cornell.2025-10-04_00-38-55z.35samp.png differ
diff --git a/IMAGES/cornell.2025-10-08_00-20-55z.2110samp.png b/IMAGES/cornell.2025-10-08_00-20-55z.2110samp.png
new file mode 100644
index 000000000..31d482077
Binary files /dev/null and b/IMAGES/cornell.2025-10-08_00-20-55z.2110samp.png differ
diff --git a/IMAGES/cornell.2025-10-08_00-24-24z.2169samp.png b/IMAGES/cornell.2025-10-08_00-24-24z.2169samp.png
new file mode 100644
index 000000000..866ff9562
Binary files /dev/null and b/IMAGES/cornell.2025-10-08_00-24-24z.2169samp.png differ
diff --git a/IMAGES/cornell.2025-10-08_00-29-14z.2109samp.png b/IMAGES/cornell.2025-10-08_00-29-14z.2109samp.png
new file mode 100644
index 000000000..4df8ee79e
Binary files /dev/null and b/IMAGES/cornell.2025-10-08_00-29-14z.2109samp.png differ
diff --git a/IMAGES/cornell.2025-10-08_00-31-15z.2687samp.png b/IMAGES/cornell.2025-10-08_00-31-15z.2687samp.png
new file mode 100644
index 000000000..228243c2c
Binary files /dev/null and b/IMAGES/cornell.2025-10-08_00-31-15z.2687samp.png differ
diff --git a/IMAGES/dof_comparison.2025-10-08_00-10-07z.5000samp.png b/IMAGES/dof_comparison.2025-10-08_00-10-07z.5000samp.png
new file mode 100644
index 000000000..145af0fd0
Binary files /dev/null and b/IMAGES/dof_comparison.2025-10-08_00-10-07z.5000samp.png differ
diff --git a/IMAGES/dof_comparison.2025-10-08_00-13-16z.5000samp.png b/IMAGES/dof_comparison.2025-10-08_00-13-16z.5000samp.png
new file mode 100644
index 000000000..0502adeb4
Binary files /dev/null and b/IMAGES/dof_comparison.2025-10-08_00-13-16z.5000samp.png differ
diff --git a/IMAGES/dof_test.2025-09-24_00-05-14z.304samp.png b/IMAGES/dof_test.2025-09-24_00-05-14z.304samp.png
new file mode 100644
index 000000000..52c5ea90f
Binary files /dev/null and b/IMAGES/dof_test.2025-09-24_00-05-14z.304samp.png differ
diff --git a/IMAGES/dof_test.2025-09-24_00-06-24z.291samp.png b/IMAGES/dof_test.2025-09-24_00-06-24z.291samp.png
new file mode 100644
index 000000000..c66ce78ab
Binary files /dev/null and b/IMAGES/dof_test.2025-09-24_00-06-24z.291samp.png differ
diff --git a/IMAGES/onetwentyeightBH.2025-10-08_03-39-12z.4079samp.png b/IMAGES/onetwentyeightBH.2025-10-08_03-39-12z.4079samp.png
new file mode 100644
index 000000000..b097ac62c
Binary files /dev/null and b/IMAGES/onetwentyeightBH.2025-10-08_03-39-12z.4079samp.png differ
diff --git a/IMAGES/singleBH_V1.2025-10-03_01-06-49z.476samp.png b/IMAGES/singleBH_V1.2025-10-03_01-06-49z.476samp.png
new file mode 100644
index 000000000..62bd782b4
Binary files /dev/null and b/IMAGES/singleBH_V1.2025-10-03_01-06-49z.476samp.png differ
diff --git a/IMAGES/singleBH_V1.2025-10-03_01-23-23z.164samp.png b/IMAGES/singleBH_V1.2025-10-03_01-23-23z.164samp.png
new file mode 100644
index 000000000..a87eb525a
Binary files /dev/null and b/IMAGES/singleBH_V1.2025-10-03_01-23-23z.164samp.png differ
diff --git a/IMAGES/singleBH_V1.2025-10-03_01-32-37z.524samp.png b/IMAGES/singleBH_V1.2025-10-03_01-32-37z.524samp.png
new file mode 100644
index 000000000..4d2648623
Binary files /dev/null and b/IMAGES/singleBH_V1.2025-10-03_01-32-37z.524samp.png differ
diff --git a/IMAGES/singleBH_V1.2025-10-04_03-01-39z.593samp.png b/IMAGES/singleBH_V1.2025-10-04_03-01-39z.593samp.png
new file mode 100644
index 000000000..b94b03a2b
Binary files /dev/null and b/IMAGES/singleBH_V1.2025-10-04_03-01-39z.593samp.png differ
diff --git a/IMAGES/singleBH_V1.2025-10-04_03-03-55z.950samp.png b/IMAGES/singleBH_V1.2025-10-04_03-03-55z.950samp.png
new file mode 100644
index 000000000..c50af5527
Binary files /dev/null and b/IMAGES/singleBH_V1.2025-10-04_03-03-55z.950samp.png differ
diff --git a/IMAGES/singleBH_V1.2025-10-05_16-53-14z.97samp.png b/IMAGES/singleBH_V1.2025-10-05_16-53-14z.97samp.png
new file mode 100644
index 000000000..ceddcc633
Binary files /dev/null and b/IMAGES/singleBH_V1.2025-10-05_16-53-14z.97samp.png differ
diff --git a/IMAGES/singleBH_V1.2025-10-05_17-53-17z.661samp.png b/IMAGES/singleBH_V1.2025-10-05_17-53-17z.661samp.png
new file mode 100644
index 000000000..501e96832
Binary files /dev/null and b/IMAGES/singleBH_V1.2025-10-05_17-53-17z.661samp.png differ
diff --git a/IMAGES/singleBH_V1.2025-10-05_17-56-29z.437samp.png b/IMAGES/singleBH_V1.2025-10-05_17-56-29z.437samp.png
new file mode 100644
index 000000000..2a081091a
Binary files /dev/null and b/IMAGES/singleBH_V1.2025-10-05_17-56-29z.437samp.png differ
diff --git a/IMAGES/singleBH_V1.2025-10-07_23-42-59z.2913samp.png b/IMAGES/singleBH_V1.2025-10-07_23-42-59z.2913samp.png
new file mode 100644
index 000000000..17005036a
Binary files /dev/null and b/IMAGES/singleBH_V1.2025-10-07_23-42-59z.2913samp.png differ
diff --git a/IMAGES/space_scene_bh.2025-10-03_03-30-06z.221samp.png b/IMAGES/space_scene_bh.2025-10-03_03-30-06z.221samp.png
new file mode 100644
index 000000000..33660da86
Binary files /dev/null and b/IMAGES/space_scene_bh.2025-10-03_03-30-06z.221samp.png differ
diff --git a/IMAGES/space_scene_bh.2025-10-03_03-31-17z.407samp.png b/IMAGES/space_scene_bh.2025-10-03_03-31-17z.407samp.png
new file mode 100644
index 000000000..5ae1281c0
Binary files /dev/null and b/IMAGES/space_scene_bh.2025-10-03_03-31-17z.407samp.png differ
diff --git a/IMAGES/space_scene_bh.2025-10-03_03-33-30z.156samp.png b/IMAGES/space_scene_bh.2025-10-03_03-33-30z.156samp.png
new file mode 100644
index 000000000..ba2b5d611
Binary files /dev/null and b/IMAGES/space_scene_bh.2025-10-03_03-33-30z.156samp.png differ
diff --git a/IMAGES/space_scene_bh.2025-10-03_12-57-05z.119samp.png b/IMAGES/space_scene_bh.2025-10-03_12-57-05z.119samp.png
new file mode 100644
index 000000000..d4eb9e261
Binary files /dev/null and b/IMAGES/space_scene_bh.2025-10-03_12-57-05z.119samp.png differ
diff --git a/README.md b/README.md
index 110697ce7..54b2dd3a0 100644
--- a/README.md
+++ b/README.md
@@ -3,11 +3,284 @@ CUDA Path Tracer
**University of Pennsylvania, CIS 565: GPU Programming and Architecture, Project 3**
-* (TODO) YOUR NAME HERE
-* Tested on: (TODO) Windows 22, i7-2222 @ 2.22GHz 22GB, GTX 222 222MB (Moore 2222 Lab)
+* Lewis Ghrist
+* [Personal Website](https://siwel-cg.github.io/siwel.cg_websiteV1/index.html#home), [LinkedIn](https://www.linkedin.com/in/lewis-ghrist-4b1b3728b/)
+* Tested on: Windows 11, AMD Ryzen 9 5950X 16-Core Processor, 64GB, NVIDIA GeForce RTX 3080 10GB
-### (TODO: Your README)
+# Renders
+
+
+
+
+
+
+
+
+
+
+
+
-*DO NOT* leave the README to the last minute! It is a crucial part of the
-project, and we will not be able to grade you without a good README.
+# Overview
+First, a quick introduction to **Path tracing**. Path tracing is a physically-based rendering technique that simulates light transport by tracing rays backward from the camera into the scene. Each ray bounces off surfaces, accumulating color and lighting information until it either hits a light source or is terminated. By averaging many randomly sampled paths per pixel, the algorithm converges to a photorealistic image with accurate global illumination, soft shadows, and complex light interactions.
+
+### GPU Implementation
+
+This path tracer uses a **wavefront architecture** optimized for GPU parallelism. Instead of assigning each thread a complete path (which would cause divergence as paths terminate at different times), each thread processes a single path segment-one bounce at a time. This allows for other optimizations such as stream compaction and material sorting, which I will explain later on. The general process of path tracing is:
+
+1. Rays are cast from the camera with initial energy (throughput) of 1.0
+2. All active rays are evaluated in parallel for their current bounce
+3. Upon intersection, the ray's throughput is multiplied by the surface's BSDF response
+4. Surviving rays continue to the next bounce iteration
+
+This approach maintains high GPU occupancy by keeping threads synchronized at each bounce level, avoiding the warp divergence that would occur if different threads were at different depths in their paths.
+For a deeper dive into path tracing, see [PBRT](https://www.pbr-book.org/3ed-2018/Light_Transport_I_Surface_Reflection/Path_Tracing) or [Ray Tracing in One Weekend](https://raytracing.github.io/).
+
+---
+
+In this project, various visual and performace improving features were implemented including:
+### Core Rendering
+- **Physically-Based Materials**: Diffuse and mirror BSDFs with stochastic roughness-based blending
+- **Stochastic Anti-Aliasing**: Randomized subpixel sampling for smooth edges
+- **Environment Mapping**: HDR skybox lighting with spherical coordinate sampling
+
+### Advanced Effects
+- **Black Hole Gravitational Lensing**: Physically accurate light bending with procedural accretion disk (see below)
+- **Depth of Field**: Thin lens camera model with configurable focal distance and aperture size
+- **Bloom Post-Processing**: Perceptual luminance-based glow for bright light sources
+
+### Performance Optimizations
+- **BVH Acceleration**: Custom bounding volume hierarchy for fast ray-mesh intersection
+- **Stream Compaction**: Automatic culling of terminated ray paths to maintain GPU efficiency
+- **Material Sorting**: Coherent BSDF evaluation through dynamic ray reordering
+
+### Pipeline
+- **Custom OBJ Loader**: Direct obj mesh import supporting positions and normals
+
+# Featurs
+## Black Hole Gravitational Lensing
+
+Path tracing typically assumes light travels in perfectly straight lines-and for the most part, that's accurate. Even phenomena like refraction that *seem* to warp light are really just straight-line segments through different media. Black holes, however, are a dramatic exception. Their immense mass distorts spacetime itself, bending the paths of light rays in ways that can't be modeled with simple geometry.
+
+This seemed like the perfect challenge for a path tracer.
+
+### The Approach
+
+The concept is straightforward: when a ray intersects the black hole's influence sphere, instead of tracing straight ahead, we simulate the ray's trajectory as it curves under gravitational acceleration toward the center-much like simulating a particle under Newtonian gravity (because fundamentally, that's what's happening to the photon).
+
+During each integration step:
+- If the ray passes within a minimum radius (the event horizon), it's absorbed-we zero out the path's energy, leaving the pixel black
+- If it exceeds a maximum radius, it's escaped the gravitational field-we resume standard straight-line ray marching until the next scene intersection
+
+Simple, right?
+
+### The Physics
+
+The challenge: *how strong is the gravitational pull?* I'm not a physicist, and general relativity is... complicated. Fortunately, much smarter people have done the heavy lifting. The implementation is based on [this excellent article](https://rantonels.github.io/starless/) by rantonels, which derives a surprisingly elegant formula for the acceleration experienced by light near a Schwarzschild (non-rotating, uncharged) black hole.
+
+With this acceleration equation in hand, I implemented an RK4 integrator for numerical stability and efficiency. The result: light rays that genuinely curve through spacetime around our black hole sphere.
+
+
+
+
+
+
+### Accretion Disk
+As you can see from the renders above, there is a bit more to a black hole than just light bending. Real black holes (at least the ones we can see) have this disk of glowing gass and debris spiraling around and into it. This glowing disk is what makes the gravitational lensing visible: light from the back of the disk bends over and around the black hole, creating the iconic "halo" effect. Simulating actual clouds of volumes would be another level of complexity I neither had the time nor the need to do. Instead, we can get a somewhat convincing result by faking this disk with a noise. If our stepped ray passes through the accretion disk plane, we can use that position to sample a noise and shade the ray accordingly.
+
+
+
+
+
+
+
+### Implementation
+So how does this all fit into our path tracer setup? As mentioned, I reuse the sphere intersection setup I had and treat the black hole as a material with a few key parameters: an RGB color channel, emittance, inner radius (event horizon), and outer radius (influence boundary). When a ray hits an object with this material, instead of performing standard BSDF evaluation, we hand it off to a specialized `blackHoleRay()` function that handles the curved spacetime integration.
+
+Starting from the intersection point, we initialize the ray's position relative to the black hole center and march it forward using RK4 integration. At each step, we update both position and velocity based on the gravitational acceleration formula from the Schwarzschild metric:
+
+$$\mathbf{a} = \frac{-3Mh^2}{|\mathbf{r}|^5} \mathbf{r} \cdot w$$
+
+Where $M$ is the black hole mass, $h^2$ is the squared angular momentum (impact parameter), and $w$ is a windowing function that smoothly attenuates the force near the boundaries. The time step adapts based on the local curvature-smaller steps near the event horizon, larger steps farther out.
+
+**Termination Conditions:**
+
+During integration, we check for three outcomes:
+
+1. **Event Horizon Capture**: Rays that get too close to the center zero out throughput and terminate.
+2. **Accretion Disk Intersection**: Rays that cross the equatorial plane within disk bounds sample the noise function for emission (see below).
+3. **Escape**: Rays that exit the outer radius moving outward return to normal path tracing.
+
+**Accretion Disk Sampling:**
+
+After each step our ray inside the gravitation field takes, I check if it passes our accretion disk's plane. If so, I find where between the current position and the last position it crosses this plane. Using that 2D coordinate, I sample a simple perlin noise function and then swirl the result based on the radius from the center very similarly to the technique I used in this past black hole project: [black hole shader work](https://siwel-cg.github.io/siwel.cg_websiteV1/projects/BlackHole.html). This gives the spiralling look without having to incorperate any actual motion into the black hole math. This noise is combined with a fall off of the radius to get a final value which I use to stochastically determine if a ray should stop and apply the emmited color to the path or continue going, passing through the accretion disk. This stochastic approach means some rays pass through the disk while others are absorbed, naturally creating the wispy, turbulent appearance of the accretion material. Although slightly ineficient, since to get a smooth, converged opacity you need to trace many rays, with this wavefront setup, this was the only way I could think of to do any sort of partial alpha effect.
+
+The best part about doing an accretion disk procedurally is that it is really easy to control the final visual output of the black hole. By varying some paramters in the noise functions, I can get different swirl intensities and densities of the disk. The nice thing about this approach is its modularity. From the path tracer's perspective, hitting a black hole is just another material evaluation, it updates the ray state and returns. Rays that escape continue bouncing through the scene normally, allowing the black hole to seamlessly composite with standard geometry and materials.
+
+**A Quick Note On Efficency**
+
+I will have more details and FPS analysis later on, but it should be intuitive that marching along a path is significantly slower than a simple mirror or diffuse bouce computation. This means that the treads for paths going through the black hole take longer than the threads who don't. At each wavefront iteration we need to sync up all the threads which means those quicker threads will have to wait. One optimization that helps with this is sorting by material type and making them contiguous in meory (this is part of the reason why I implemented this black holes as a material). I didn't really implement any other GPU specific optimizations for this, but one could be doing stream compaction for substep of our walk, similar to what we do for the actuall path segments. Even though, particularly for open scenes, the light distortion basically was real time, in close scenes where many paths bounce in and out of the black hole multiple times, it can have a significant performance impact. Most of my scenes and testing involved just 1 or 2 black holes, but if you have a scene with many, the same problem could occure. Using RK4 and updating time steps certainly does help with efficency, but future work could be done to take advantage of the parallel architecture even more for better results.
+
+# Visual Improvements
+Besides this flashy black hole shader, I implemented a few other featurs which help to enhanse the effect of the black hole or overall just allows for more interesting visuals. The first being Bloom.
+
+### Bloom
+Bloom is a post process effect which adds an artificial glow to parts of the image which pass a certain birightness threshold. We naturally get this effect due to light bouncing around in our eye, but in a simulated world without an actual participating media for the light rays to travel to, this effect doesn't happen. But we can fake it in post. After the full image calculation has be run and we average the light values for all the rays of an iteration, we then do pass on each pixel and determine if it passes this light threshold, keeping only the ones that pass. From there, to get the glow effect, we blur this light filter using a Gaussian blur. In my implementation I used a 21x21 kernel, but the strength of the blur can be adjusted as needed. This blurred pass is then added back to our original image, giving it an angelic glowing effect. Particularly for the black hole, this makes quite the difference:
+
+
+
+
+
+
+### Environment Mapping
+
+To light scenes with realistic outdoor lighting, as well as test my black hole distortin, I implemented HDR environment map support. An environment map is essentially an image wrapped around the scene at infinite distance, providing both illumination and background imagery. When a ray fails to intersect any geometry in the scene, rather than returning black, we sample the environment map based on the ray's direction. The ray direction (a 3D vector) is converted to spherical coordinates theta (azimuthal angle) and phi (polar angle) which map to UV coordinates on the environment texture:
+
+$$u = \frac{1}{2} + \frac{\arctan2(d_z, d_x)}{2\pi}, \quad v = \frac{1}{2} - \frac{\arcsin(d_y)}{\pi}$$
+
+Where $\mathbf{d}$ is the normalized ray direction. This spherical mapping allows a 2D image to represent all possible incoming light directions.
+
+The environment map integrates naturally into the path tracing pipeline: it's simply another potential light source. Rays accumulate color from environment lighting just like they would from any emissive surface, and the map's contribution is weighted by the path's current throughput.
+
+
+
+
+
+
+### Thin Lense Depth Of Field
+
+Real cameras have finite apertures, creating a depth of field effect where objects at the focal distance appear sharp while objects closer or farther away become progressively blurred. I implemented this using a thin lens camera model. Unlike a pinhole camera where every ray passes through a single point (keeping everything in perfect focus), a thin lens has an aperture with non-zero radius. Rays originating from different points on the lens converge at the focal plane but diverge elsewhere, creating blur.
+
+
+
+
+
+
+
+The implementation samples random points on the circular lens aperture using concentric disk sampling, then adjusts each ray's direction so it passes through the same point on the focal plane that the original ray (from the lens center) would have hit. Over many frames, rays from different lens positions average together points at the focal distance receive consistent samples and appear sharp, while points at other depths receive divergent samples, creating blur proportional to their distance from the focal plane. The effect is controlled by two parameters: lens radius (aperture size, where larger means stronger blur) and focal distance (which depth appears sharp).
+
+
+
+
+
+
+### Stochastic Anti-Aliasing
+
+Similarly to how we scattered ray origins across the lens aperture to achieve depth of field, we can apply the same stochastic sampling principle to eliminate aliasing. Instead of casting rays through the exact center of each pixel, we jitter the ray origin randomly within the pixel's area. Without this, rendered images suffer from jagged edges where object boundaries don't align perfectly with pixel centers causing a "staircase" like artifact. Each frame uses a different random offset within the pixel, so over many iterations the samples average across the entire pixel area. Edges that partially cover a pixel receive proportionally mixed colors, naturally producing the correct blended color. This approach requires no special edge detection or additional samples per frame, unlike what you would need for a rasterize. The anti-aliasing emerges automatically from the same Monte Carlo integration that drives the path tracing itself.
+
+
+# Performance Improvements
+
+Path tracing is computationally expensive, and even with the parallel power of a GPU, without good thread utilization, performence can still be slow. Several optimizations were crucial to achieving interactive frame rates. The first and most important one was a Bounding Volume Heiarchy (BVH) which allowed for OBJ mesh loading.
+
+### BVH and OBJs
+The goal of BVH is to reducce the number of intersection tests you need to do for a ray. In a niave path tracer implementation, for each ray, you check every single object in your scene to see if the ray hits that object. For anything more complicated than a few simple pieces of geometry, this takes forever. In order to do any type of custom mesh loading, BVH or some other type of acceleration structure was crucial.
+
+A BVH solves the intersection scaling problem by organizing geometry into a tree of nested bounding boxes. The key insight is that if a ray doesn't hit a bounding box, it can't possibly hit any of the geometry inside that box, allowing us to skip entire branches of the tree and taking our intersection test time from $O(N)$ to $O(log(N))$. First, the hierarchy is built recursively on the CPU. For agiven node, we compute the overall bounding box of that node, the partition each internal piece of geometry based on its centroid and the midpoint of the longest axis of our bounding box. This creates two sets of geometry, one on the "left" of the midpoint and one on the "right". These then go on to become their own nodes and so on until we reach a minimum size limit and we get a leaf node.
+
+Before we build this BVH tree, we first need to load in our OBJ triangle meshes into our geometry array. For simplicity, I used the [tinyobjloader](https://github.com/tinyobjloader/tinyobjloader/tree/release) library which automatically handles reading in an OBJ and converting it into triangle positions and normals with the correct indices. With these triangles I precompute the centroid positions, and finally store the actual geometry into the goemetry array. After which, we build our BVH tree.
+
+This tree is then sent to the GPU via a linearized tree structure rather than pointer-based nodes. Nodes are stored in a flat array with children accessed via index offsets and geometry stored as start and end indices in our Geometry array. This provides better cache coherence on the GPU, where pointer chasing is expensive. During rendering, ray-BVH intersection uses a stack-based traversal. Starting at the root, we test the ray against the node's AABB. If it misses, we pop back up the tree. If it hits and the node is internal, we push both children onto the stack. If it hits a leaf node, we test against all triangles in that leaf. The closest intersection found across all tested triangles is returned. This allows rays to skip vast portions of the scene transforming render times for complext models from minutes per frame to interactive rates.
+
+
+
+
+
+
+### Stream Compaction
+As I have mentioned a few times now, one area for optimization is in culling "dead" rays so that we don't use threads to calculate nothing. For this, we can use stream compaction. Stream compaction is a parallel way of doing a linear search through an array and removing unwanted elements while shifting all the other elements downwards so they are continuous in the array. We can take advantage of this algorithm to efficiently remove these useless path segments each bounce. For this, I use `thrust::partition` to separate paths into two groups: those still alive (`remainingBounces > 0`) and those that have finished. The partition operation is stable and efficient, rearranging the path array in-place so all active paths are packed at the front. We then update the path count to reflect only the active paths, and subsequent bounces operate on this smaller buffer. For a closed scene with only a few lights, this will have a minimum effect. However, the real advantage comes when you have a very open scene. Since many of the rays will go of into empty space and terminate just after the first bounce, at each iteration, many of our rays will beocme usless. By dynamically culling them, we can drastically reduce the amount of wasted kernel calls resulting in much faster renders.
+
+### Material Sorting
+The last bit of optimization we can do, is sorting the path segments based on the materials they hit. What we want is for each warp to execute the same instructions coherently. When paths hit different material types and evaluate different BSDFs, they diverge and some threads execute, for example, diffuse shading code while others execute specular reflections, This warp divergence forces the GPU to serialize execution, dramatically reducing throughput. Particularly, as I mentioned, for my black hole ray stepping material. Material sorting addresses this by reordering paths before shading so that rays hitting the same material type are grouped together. I use `thrust::sort_by_key` with the material ID as the key and the path segment as the value. After sorting, all paths evaluating diffuse materials execute consecutively, followed by all specular paths, then black hole paths, and so on. Threads within each warp now execute the same BSDF code path, eliminating divergence and improving memory access coherence since similar materials often have similar memory layouts.
+
+# Performance Analysis
+### Stream Compaction
+As mentioned, stream compaction can significantly impact render performance particularly on very open scenes. As such, I compared the average frame rate over on both open and closed scenes with minimal objects in them and with a bounce limit of 24. The data shows this improvement with the stream compaction having a marginal increase in FPS for the close scene, and a much larger increase for the open scene.
+
+
+
+
+
+### BVH
+BVH is what made object loading possible. Even though for small scenes, the overhead of constructing the tree is not worth it, as I found for my implementation, for anything over about 200 objects, is not even viable. For the larger models, the render would instantly crash on the first iteration of bounces. Again, across various scenes with differing numbers of objects, I took the average FPS over a 30 second window. The scenes were partially open, so stream compaction had some effect, however the majority of the "heavy lifting" was done by the BVH.
+
+
+
+
+
+### Black Holes
+Since a black hole thread takes much longer than a regular thread, I also tested scenes with multiple black holes. For simplicity, I had just black holes in a completely open scene so that random bounces from diffuse surfaces wouldn't alter any results. Although, as expected, the results get worse the more black holes there are, I was plesently supprised. Even with 128 black holes I was still getting around 44 fps, which not too bad.
+
+
+
+
+
+Here is what it looked like with 128 black holes:
+
+
+
+
+
+### Material Sorting
+Although in some cases, sorting the segments based on terial would help reduce warp divergence and speed up render times, I found that my scenes never had enough materials to make this worth while. In fact, across all my scenes, there was a consistent drop in performance when I did sort the paths. My focus for this path tracer wasn't a vast amount of nice PBR materials with many different effects so although I implemented it as a future optimization for when I do increase the number of material types, for now, I found its more harmful than helpful.
+
+---
+
+# Conclusion
+Overall I was very happy with how this project turned out. The light bending and black holes, in my opinion, look really nice and render pretty quickly. As for some of the other implemented features, there is always room for improvement. First off, some things are kind of buggy and there are a few messy parts of the code I want to go back and clean up. Additionally I would like to add support for more types of materials including Transmisive materials, Sub Surface Scattering, and maybe volumes. Some better UI and parameters would also be really helpful for loading models, scenes, and environment maps. As a final send off, here are some more random renders I took through out the process of working on this project.
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+### Bloopers
+These are just some wild renderes I got while trying to implement some of these features.
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+# References
+- https://henrikdahlberg.github.io/2016/08/23/stream-compaction.html
+- https://nvidia.github.io/cccl/thrust/api/group__stream__compaction_1gaf01d45b30fecba794afae065d625f94f.html
+- https://jacco.ompf2.com/2022/04/13/how-to-build-a-bvh-part-1-basics/
+- http://spiro.fisica.unipd.it/~antonell/schwarzschild/
+- https://rantonels.github.io/starless/
+- https://web.mit.edu/10.001/Web/Course_Notes/Differential_Equations_Notes/node5.html
+- https://blog.seanholloway.com/2022/03/13/visualizing-black-holes-with-general-relativistic-ray-tracing/
+- https://learnopengl.com/Advanced-Lighting/Bloom
+- https://github.com/tinyobjloader/tinyobjloader/tree/release
+- https://free3d.com/3d-model/hand-v3--902450.html
+- https://www.spacespheremaps.com/hdr-spheremaps/
+- https://pbr-book.org/3ed-2018/Camera_Models/Realistic_Cameras
\ No newline at end of file
diff --git a/enviMaps/01d69222-cd93-4da2-85ad-95e02318777f.png b/enviMaps/01d69222-cd93-4da2-85ad-95e02318777f.png
new file mode 100644
index 000000000..4e8c81d6b
Binary files /dev/null and b/enviMaps/01d69222-cd93-4da2-85ad-95e02318777f.png differ
diff --git a/enviMaps/GPT_Galaxy_V2.png b/enviMaps/GPT_Galaxy_V2.png
new file mode 100644
index 000000000..23e0f9f25
Binary files /dev/null and b/enviMaps/GPT_Galaxy_V2.png differ
diff --git a/enviMaps/GPT_Galaxy_V3.png b/enviMaps/GPT_Galaxy_V3.png
new file mode 100644
index 000000000..eda07f568
Binary files /dev/null and b/enviMaps/GPT_Galaxy_V3.png differ
diff --git a/enviMaps/GPT_Galaxy_V4.png b/enviMaps/GPT_Galaxy_V4.png
new file mode 100644
index 000000000..99a77301a
Binary files /dev/null and b/enviMaps/GPT_Galaxy_V4.png differ
diff --git a/enviMaps/GPT_Galaxy_V5.png b/enviMaps/GPT_Galaxy_V5.png
new file mode 100644
index 000000000..37e73fd93
Binary files /dev/null and b/enviMaps/GPT_Galaxy_V5.png differ
diff --git a/enviMaps/HDR_blue_nebulae-1.hdr b/enviMaps/HDR_blue_nebulae-1.hdr
new file mode 100644
index 000000000..f34b25d78
Binary files /dev/null and b/enviMaps/HDR_blue_nebulae-1.hdr differ
diff --git a/enviMaps/HDR_galactic_plane_hazy_nebulae.hdr b/enviMaps/HDR_galactic_plane_hazy_nebulae.hdr
new file mode 100644
index 000000000..1e1e34f67
Binary files /dev/null and b/enviMaps/HDR_galactic_plane_hazy_nebulae.hdr differ
diff --git a/enviMaps/HDR_rich_multi_nebulae_2.hdr b/enviMaps/HDR_rich_multi_nebulae_2.hdr
new file mode 100644
index 000000000..3cd165daf
Binary files /dev/null and b/enviMaps/HDR_rich_multi_nebulae_2.hdr differ
diff --git a/enviMaps/Milkway_Image_V2.png b/enviMaps/Milkway_Image_V2.png
new file mode 100644
index 000000000..c699ce77a
Binary files /dev/null and b/enviMaps/Milkway_Image_V2.png differ
diff --git a/enviMaps/MilkyWay.jpg b/enviMaps/MilkyWay.jpg
new file mode 100644
index 000000000..9689a6c99
Binary files /dev/null and b/enviMaps/MilkyWay.jpg differ
diff --git a/enviMaps/citrus_orchard_road_puresky_4k.hdr b/enviMaps/citrus_orchard_road_puresky_4k.hdr
new file mode 100644
index 000000000..dfd5ad511
Binary files /dev/null and b/enviMaps/citrus_orchard_road_puresky_4k.hdr differ
diff --git a/enviMaps/earthlike_planet.hdr b/enviMaps/earthlike_planet.hdr
new file mode 100644
index 000000000..729289b19
Binary files /dev/null and b/enviMaps/earthlike_planet.hdr differ
diff --git a/enviMaps/rogland_clear_night_4k.hdr b/enviMaps/rogland_clear_night_4k.hdr
new file mode 100644
index 000000000..baf4cd542
Binary files /dev/null and b/enviMaps/rogland_clear_night_4k.hdr differ
diff --git a/enviMaps/the_sky_is_on_fire_4k.hdr b/enviMaps/the_sky_is_on_fire_4k.hdr
new file mode 100644
index 000000000..7d729010d
Binary files /dev/null and b/enviMaps/the_sky_is_on_fire_4k.hdr differ
diff --git a/scenes/BHTestSceene_V2.json b/scenes/BHTestSceene_V2.json
new file mode 100644
index 000000000..6a00b2297
--- /dev/null
+++ b/scenes/BHTestSceene_V2.json
@@ -0,0 +1,68 @@
+{
+ "Materials": {
+ "light": {
+ "TYPE": "Emitting",
+ "RGB": [1.0, 1.0, 1.0],
+ "EMITTANCE": 5.0,
+ "ROUGHNESS": 1.0
+ },
+ "diffuse_white": {
+ "TYPE": "Diffuse",
+ "RGB": [0.98, 0.98, 0.98],
+ "ROUGHNESS": 1.0
+ },
+ "diffuse_red": {
+ "TYPE": "Diffuse",
+ "RGB": [0.85, 0.35, 0.35],
+ "ROUGHNESS": 1.0
+ },
+ "diffuse_green": {
+ "TYPE": "Diffuse",
+ "RGB": [0.35, 0.85, 0.35],
+ "ROUGHNESS": 1.0
+ },
+ "diffuse_blue": {
+ "TYPE": "Diffuse",
+ "RGB": [0.35, 0.35, 0.85],
+ "ROUGHNESS": 1.0
+ },
+ "diffuse_yellow": {
+ "TYPE": "Diffuse",
+ "RGB": [0.85, 0.85, 0.35],
+ "ROUGHNESS": 1.0
+ },
+ "blackhole": {
+ "TYPE": "BlackHole",
+ "RGB": [0.0, 0.0, 0.0],
+ "EMITTANCE": 0.0,
+ "ROUGHNESS": 1.0,
+ "INNERRAD": 1.0,
+ "OUTERRAD": 8.0
+ }
+ },
+
+ "Camera": {
+ "RES": [800, 800],
+ "FOVY": 45.0,
+ "ITERATIONS": 5000,
+ "DEPTH": 32,
+ "FILE": "bvh_grid_test_bh",
+ "EYE": [0.0, 8.0, 12.0],
+ "LOOKAT": [0.0, 3.0, 0.0],
+ "UP": [0.0, 1.0, 0.0]
+ },
+
+ "Objects": [
+ { "TYPE": "sphere", "MATERIAL": "diffuse_red", "TRANS": [-3.0, 2.0, -3.0], "ROTAT": [0.0,0.0,0.0], "SCALE": [1.0,1.0,1.0] },
+ { "TYPE": "sphere", "MATERIAL": "diffuse_green", "TRANS": [ 0.0, 2.0, -3.0], "ROTAT": [0.0,0.0,0.0], "SCALE": [1.0,1.0,1.0] },
+ { "TYPE": "sphere", "MATERIAL": "diffuse_blue", "TRANS": [ 3.0, 2.0, -3.0], "ROTAT": [0.0,0.0,0.0], "SCALE": [1.0,1.0,1.0] },
+ { "TYPE": "sphere", "MATERIAL": "diffuse_yellow", "TRANS": [-3.0, 2.0, 0.0], "ROTAT": [0.0,0.0,0.0], "SCALE": [1.0,1.0,1.0] },
+
+ { "TYPE": "sphere", "MATERIAL": "blackhole", "TRANS": [ 0.0, 2.0, 0.0], "ROTAT": [0.0,0.0,0.0], "SCALE": [8.0,8.0,8.0] },
+
+ { "TYPE": "sphere", "MATERIAL": "diffuse_red", "TRANS": [ 3.0, 2.0, 0.0], "ROTAT": [0.0,0.0,0.0], "SCALE": [1.0,1.0,1.0] },
+ { "TYPE": "sphere", "MATERIAL": "diffuse_green", "TRANS": [-3.0, 2.0, 3.0], "ROTAT": [0.0,0.0,0.0], "SCALE": [1.0,1.0,1.0] },
+ { "TYPE": "sphere", "MATERIAL": "diffuse_blue", "TRANS": [ 0.0, 2.0, 3.0], "ROTAT": [0.0,0.0,0.0], "SCALE": [1.0,1.0,1.0] },
+ { "TYPE": "sphere", "MATERIAL": "diffuse_yellow", "TRANS": [ 3.0, 2.0, 3.0], "ROTAT": [0.0,0.0,0.0], "SCALE": [1.0,1.0,1.0] }
+ ]
+}
diff --git a/scenes/blackhole_scene.json b/scenes/blackhole_scene.json
new file mode 100644
index 000000000..3d7900aea
--- /dev/null
+++ b/scenes/blackhole_scene.json
@@ -0,0 +1,149 @@
+{
+ "Materials": {
+ "blackhole": {
+ "TYPE": "BlackHole",
+ "RGB": [0.15, 0.05, 0.25],
+ "EMITTANCE": 9.0,
+ "ROUGHNESS": 1.0,
+ "INNERRAD": 2.0,
+ "OUTERRAD": 24.0
+ },
+ "light_blue": {
+ "TYPE": "Emitting",
+ "RGB": [0.4, 0.7, 1.0],
+ "EMITTANCE": 15.0,
+ "ROUGHNESS": 1.0
+ },
+ "light_cyan": {
+ "TYPE": "Emitting",
+ "RGB": [0.3, 0.9, 0.9],
+ "EMITTANCE": 14.0,
+ "ROUGHNESS": 1.0
+ },
+ "light_purple": {
+ "TYPE": "Emitting",
+ "RGB": [0.7, 0.4, 1.0],
+ "EMITTANCE": 13.0,
+ "ROUGHNESS": 1.0
+ },
+ "light_magenta": {
+ "TYPE": "Emitting",
+ "RGB": [1.0, 0.3, 0.8],
+ "EMITTANCE": 14.0,
+ "ROUGHNESS": 1.0
+ },
+ "light_pink": {
+ "TYPE": "Emitting",
+ "RGB": [1.0, 0.5, 0.7],
+ "EMITTANCE": 12.0,
+ "ROUGHNESS": 1.0
+ },
+ "light_white": {
+ "TYPE": "Emitting",
+ "RGB": [0.9, 0.95, 1.0],
+ "EMITTANCE": 16.0,
+ "ROUGHNESS": 1.0
+ },
+ "specular_chrome": {
+ "TYPE": "Specular",
+ "RGB": [0.98, 0.98, 0.98],
+ "ROUGHNESS": 0.0
+ },
+ "specular_blue": {
+ "TYPE": "Specular",
+ "RGB": [0.6, 0.75, 0.95],
+ "ROUGHNESS": 0.0
+ },
+ "specular_purple": {
+ "TYPE": "Specular",
+ "RGB": [0.75, 0.6, 0.9],
+ "ROUGHNESS": 0.0
+ },
+ "specular_silver": {
+ "TYPE": "Specular",
+ "RGB": [0.85, 0.88, 0.95],
+ "ROUGHNESS": 0.05
+ },
+ "glossy_blue": {
+ "TYPE": "Diffuse",
+ "RGB": [0.3, 0.5, 0.8],
+ "ROUGHNESS": 0.3
+ },
+ "glossy_purple": {
+ "TYPE": "Diffuse",
+ "RGB": [0.6, 0.3, 0.7],
+ "ROUGHNESS": 0.3
+ },
+ "glossy_cyan": {
+ "TYPE": "Diffuse",
+ "RGB": [0.3, 0.7, 0.8],
+ "ROUGHNESS": 0.4
+ },
+ "glossy_magenta": {
+ "TYPE": "Diffuse",
+ "RGB": [0.7, 0.3, 0.6],
+ "ROUGHNESS": 0.3
+ },
+ "glossy_indigo": {
+ "TYPE": "Diffuse",
+ "RGB": [0.4, 0.3, 0.7],
+ "ROUGHNESS": 0.35
+ }
+ },
+
+ "Camera": {
+ "RES": [800, 800],
+ "FOVY": 45.0,
+ "ITERATIONS": 5000,
+ "DEPTH": 32,
+ "FILE": "blackhole_bvh_test",
+ "EYE": [0.0, 10.0, 30.0],
+ "LOOKAT": [0.0, 5.0, 0.0],
+ "UP": [0.0, 1.0, 0.0]
+ },
+
+ "Objects": [
+ { "TYPE": "sphere", "MATERIAL": "blackhole", "TRANS": [0.0, 5.0, 0.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [24.0, 24.0, 24.0] },
+
+ { "TYPE": "sphere", "MATERIAL": "light_white", "TRANS": [18.0, 14.0, 2.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [0.8, 0.8, 0.8] },
+ { "TYPE": "sphere", "MATERIAL": "light_blue", "TRANS": [-16.0, 12.0, 8.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [0.7, 0.7, 0.7] },
+ { "TYPE": "sphere", "MATERIAL": "light_cyan", "TRANS": [12.0, 16.0, -10.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [0.7, 0.7, 0.7] },
+ { "TYPE": "sphere", "MATERIAL": "light_purple", "TRANS": [-14.0, 10.0, -12.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [0.8, 0.8, 0.8] },
+ { "TYPE": "sphere", "MATERIAL": "light_magenta", "TRANS": [10.0, 8.0, 15.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [0.7, 0.7, 0.7] },
+ { "TYPE": "sphere", "MATERIAL": "light_pink", "TRANS": [-8.0, 18.0, 6.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [0.6, 0.6, 0.6] },
+ { "TYPE": "sphere", "MATERIAL": "light_cyan", "TRANS": [15.0, 6.0, -8.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [0.7, 0.7, 0.7] },
+ { "TYPE": "sphere", "MATERIAL": "light_blue", "TRANS": [-10.0, 15.0, 14.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [0.8, 0.8, 0.8] },
+
+ { "TYPE": "sphere", "MATERIAL": "specular_chrome", "TRANS": [8.0, 3.0, 8.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [2.2, 2.2, 2.2] },
+ { "TYPE": "sphere", "MATERIAL": "specular_chrome", "TRANS": [-9.0, 4.0, 7.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [2.0, 2.0, 2.0] },
+ { "TYPE": "sphere", "MATERIAL": "specular_blue", "TRANS": [7.0, 7.0, -8.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [1.9, 1.9, 1.9] },
+ { "TYPE": "sphere", "MATERIAL": "specular_purple", "TRANS": [-8.0, 2.0, -9.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [1.8, 1.8, 1.8] },
+ { "TYPE": "sphere", "MATERIAL": "specular_silver", "TRANS": [12.0, 5.0, 4.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [2.1, 2.1, 2.1] },
+ { "TYPE": "sphere", "MATERIAL": "specular_chrome", "TRANS": [-11.0, 8.0, -5.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [1.7, 1.7, 1.7] },
+ { "TYPE": "sphere", "MATERIAL": "specular_blue", "TRANS": [5.0, 11.0, 10.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [1.6, 1.6, 1.6] },
+ { "TYPE": "sphere", "MATERIAL": "specular_purple", "TRANS": [-6.0, 9.0, -11.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [1.8, 1.8, 1.8] },
+
+ { "TYPE": "cube", "MATERIAL": "specular_chrome", "TRANS": [11.0, 5.0, -6.0], "ROTAT": [0.0, 30.0, 0.0], "SCALE": [2.5, 2.5, 2.5] },
+ { "TYPE": "cube", "MATERIAL": "specular_silver", "TRANS": [-10.0, 6.0, 5.0], "ROTAT": [0.0, -40.0, 0.0], "SCALE": [2.3, 2.3, 2.3] },
+ { "TYPE": "cube", "MATERIAL": "specular_blue", "TRANS": [4.0, 10.0, -7.0], "ROTAT": [0.0, 45.0, 15.0], "SCALE": [2.2, 2.2, 2.2] },
+ { "TYPE": "cube", "MATERIAL": "specular_purple", "TRANS": [-5.0, 12.0, 8.0], "ROTAT": [0.0, -25.0, 10.0], "SCALE": [2.0, 2.0, 2.0] },
+ { "TYPE": "cube", "MATERIAL": "specular_chrome", "TRANS": [13.0, 3.0, 9.0], "ROTAT": [0.0, 55.0, -5.0], "SCALE": [2.4, 2.4, 2.4] },
+ { "TYPE": "cube", "MATERIAL": "specular_silver", "TRANS": [-12.0, 7.0, -10.0], "ROTAT": [0.0, -50.0, 8.0], "SCALE": [2.1, 2.1, 2.1] },
+
+ { "TYPE": "sphere", "MATERIAL": "glossy_blue", "TRANS": [14.0, 2.0, 3.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [1.5, 1.5, 1.5] },
+ { "TYPE": "sphere", "MATERIAL": "glossy_purple", "TRANS": [-13.0, 3.0, -4.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [1.4, 1.4, 1.4] },
+ { "TYPE": "sphere", "MATERIAL": "glossy_cyan", "TRANS": [3.0, 1.0, 14.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [1.6, 1.6, 1.6] },
+ { "TYPE": "sphere", "MATERIAL": "glossy_magenta", "TRANS": [-4.0, 13.0, 2.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [1.3, 1.3, 1.3] },
+ { "TYPE": "sphere", "MATERIAL": "glossy_indigo", "TRANS": [16.0, 4.0, -6.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [1.4, 1.4, 1.4] },
+ { "TYPE": "sphere", "MATERIAL": "glossy_blue", "TRANS": [-15.0, 5.0, 10.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [1.3, 1.3, 1.3] },
+
+ { "TYPE": "cube", "MATERIAL": "glossy_cyan", "TRANS": [9.0, 8.0, 11.0], "ROTAT": [0.0, 25.0, 10.0], "SCALE": [1.8, 1.8, 1.8] },
+ { "TYPE": "cube", "MATERIAL": "glossy_purple", "TRANS": [-10.0, 10.0, -9.0], "ROTAT": [0.0, -35.0, 5.0], "SCALE": [1.7, 1.7, 1.7] },
+ { "TYPE": "cube", "MATERIAL": "glossy_magenta", "TRANS": [6.0, 1.0, 12.0], "ROTAT": [0.0, 50.0, -10.0], "SCALE": [1.6, 1.6, 1.6] },
+
+ { "TYPE": "sphere", "MATERIAL": "specular_chrome", "TRANS": [17.0, 8.0, 1.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [1.4, 1.4, 1.4] },
+ { "TYPE": "sphere", "MATERIAL": "specular_blue", "TRANS": [-17.0, 9.0, 0.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [1.3, 1.3, 1.3] },
+ { "TYPE": "sphere", "MATERIAL": "specular_purple", "TRANS": [2.0, 17.0, 3.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [1.2, 1.2, 1.2] },
+ { "TYPE": "sphere", "MATERIAL": "specular_silver", "TRANS": [-3.0, 16.0, -5.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [1.3, 1.3, 1.3] }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/blackhole_scene_V2.json b/scenes/blackhole_scene_V2.json
new file mode 100644
index 000000000..49b8ca413
--- /dev/null
+++ b/scenes/blackhole_scene_V2.json
@@ -0,0 +1,84 @@
+{
+ "Materials": {
+ "blackhole": {
+ "TYPE": "BlackHole",
+ "RGB": [0.9, 0.4, 0.1],
+ "EMITTANCE": 3.0,
+ "ROUGHNESS": 1.0,
+ "INNERRAD": 2.0,
+ "OUTERRAD": 30.0
+ },
+ "mirror_chrome": {
+ "TYPE": "Specular",
+ "RGB": [0.98, 0.98, 0.98],
+ "ROUGHNESS": 0.0
+ },
+ "mirror_gold": {
+ "TYPE": "Specular",
+ "RGB": [1.0, 0.85, 0.5],
+ "ROUGHNESS": 0.0
+ },
+ "mirror_blue": {
+ "TYPE": "Specular",
+ "RGB": [0.7, 0.8, 0.95],
+ "ROUGHNESS": 0.0
+ }
+ },
+
+ "Camera": {
+ "RES": [1800, 800],
+ "FOVY": 40.0,
+ "ITERATIONS": 5000,
+ "DEPTH": 32,
+ "FOCALDIST": 45.0,
+ "APERTURE": 0.8,
+ "FILE": "blackhole_mirrors",
+ "EYE": [-45.0, 8.0, 0.0],
+ "LOOKAT": [0.0, 0.0, 0.0],
+ "UP": [0.0, 1.0, 0.0]
+ },
+
+ "Objects": [
+ { "TYPE": "sphere", "MATERIAL": "blackhole", "TRANS": [0.0, 0.0, 0.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [30.0, 30.0, 30.0] },
+
+ { "TYPE": "sphere", "MATERIAL": "mirror_chrome", "TRANS": [-35.0, 0.0, 0.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [5.0, 5.0, 5.0] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_chrome", "TRANS": [35.0, 0.0, 0.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [5.0, 5.0, 5.0] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_chrome", "TRANS": [0.0, 35.0, 0.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [5.0, 5.0, 5.0] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_chrome", "TRANS": [0.0, -35.0, 0.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [5.0, 5.0, 5.0] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_chrome", "TRANS": [0.0, 0.0, -35.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [5.0, 5.0, 5.0] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_chrome", "TRANS": [0.0, 0.0, 35.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [5.0, 5.0, 5.0] },
+
+ { "TYPE": "sphere", "MATERIAL": "mirror_gold", "TRANS": [-25.0, 25.0, 0.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.5, 4.5, 4.5] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_gold", "TRANS": [25.0, 25.0, 0.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.5, 4.5, 4.5] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_gold", "TRANS": [-25.0, -25.0, 0.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.5, 4.5, 4.5] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_gold", "TRANS": [25.0, -25.0, 0.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.5, 4.5, 4.5] },
+
+ { "TYPE": "sphere", "MATERIAL": "mirror_gold", "TRANS": [-25.0, 0.0, 25.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.5, 4.5, 4.5] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_gold", "TRANS": [25.0, 0.0, 25.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.5, 4.5, 4.5] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_gold", "TRANS": [-25.0, 0.0, -25.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.5, 4.5, 4.5] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_gold", "TRANS": [25.0, 0.0, -25.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.5, 4.5, 4.5] },
+
+ { "TYPE": "sphere", "MATERIAL": "mirror_gold", "TRANS": [0.0, 25.0, -25.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.5, 4.5, 4.5] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_gold", "TRANS": [0.0, 25.0, 25.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.5, 4.5, 4.5] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_gold", "TRANS": [0.0, -25.0, -25.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.5, 4.5, 4.5] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_gold", "TRANS": [0.0, -25.0, 25.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.5, 4.5, 4.5] },
+
+ { "TYPE": "sphere", "MATERIAL": "mirror_blue", "TRANS": [-30.0, 15.0, 15.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.0, 4.0, 4.0] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_blue", "TRANS": [30.0, 15.0, 15.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.0, 4.0, 4.0] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_blue", "TRANS": [-30.0, -15.0, 15.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.0, 4.0, 4.0] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_blue", "TRANS": [30.0, -15.0, 15.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.0, 4.0, 4.0] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_blue", "TRANS": [-30.0, 15.0, -15.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.0, 4.0, 4.0] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_blue", "TRANS": [30.0, 15.0, -15.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.0, 4.0, 4.0] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_blue", "TRANS": [-30.0, -15.0, -15.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.0, 4.0, 4.0] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_blue", "TRANS": [30.0, -15.0, -15.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [4.0, 4.0, 4.0] },
+
+ { "TYPE": "sphere", "MATERIAL": "mirror_chrome", "TRANS": [-18.0, 18.0, 18.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [3.5, 3.5, 3.5] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_chrome", "TRANS": [18.0, 18.0, 18.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [3.5, 3.5, 3.5] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_chrome", "TRANS": [-18.0, -18.0, 18.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [3.5, 3.5, 3.5] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_chrome", "TRANS": [18.0, -18.0, 18.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [3.5, 3.5, 3.5] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_chrome", "TRANS": [-18.0, 18.0, -18.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [3.5, 3.5, 3.5] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_chrome", "TRANS": [18.0, 18.0, -18.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [3.5, 3.5, 3.5] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_chrome", "TRANS": [-18.0, -18.0, -18.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [3.5, 3.5, 3.5] },
+ { "TYPE": "sphere", "MATERIAL": "mirror_chrome", "TRANS": [18.0, -18.0, -18.0], "ROTAT": [0.0, 0.0, 0.0], "SCALE": [3.5, 3.5, 3.5] }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/blackhole_test_scene.json b/scenes/blackhole_test_scene.json
new file mode 100644
index 000000000..a200c8bc7
--- /dev/null
+++ b/scenes/blackhole_test_scene.json
@@ -0,0 +1,164 @@
+{
+ "Materials":
+ {
+ "blackhole":
+ {
+ "TYPE":"BlackHole",
+ "RGB":[0.0,0.0,0.0],
+ "EMITTANCE":0.0,
+ "ROUGHNESS":1.0,
+ "INNERRAD": 1.0,
+ "OUTERRAD": 10.0
+ },
+ "bright_light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0, 1.0, 1.0],
+ "EMITTANCE":30.0,
+ "ROUGHNESS":1.0
+ },
+ "warm_light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0, 0.8, 0.6],
+ "EMITTANCE":20.0,
+ "ROUGHNESS":1.0
+ },
+ "cool_light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[0.6, 0.8, 1.0],
+ "EMITTANCE":18.0,
+ "ROUGHNESS":1.0
+ },
+ "accent_light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[0.9, 0.7, 1.0],
+ "EMITTANCE":15.0,
+ "ROUGHNESS":1.0
+ },
+ "diffuse_white":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.9, 0.9, 0.9],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_red":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.8, 0.3, 0.3],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_blue":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.3, 0.5, 0.8],
+ "ROUGHNESS":1.0
+ },
+ "mirror":
+ {
+ "TYPE":"Specular",
+ "RGB":[0.95, 0.95, 0.95],
+ "ROUGHNESS":0.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[800,800],
+ "FOVY":55.0,
+ "ITERATIONS":5000,
+ "DEPTH":32,
+ "FILE":"blackhole_test",
+ "EYE":[0.0,5.0,25.0],
+ "LOOKAT":[0.0,0.0,0.0],
+ "UP":[0.0,1.0,0.0]
+ },
+ "Objects":
+ [
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"blackhole",
+ "TRANS":[0.0,0.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[10.0,10.0,10.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"bright_light",
+ "TRANS":[15.0,8.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.5,2.5,2.5]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"warm_light",
+ "TRANS":[-12.0,6.0,8.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.0,2.0,2.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"cool_light",
+ "TRANS":[8.0,-5.0,12.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.2,2.2,2.2]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"accent_light",
+ "TRANS":[-10.0,-4.0,-10.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.8,1.8,1.8]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"bright_light",
+ "TRANS":[0.0,18.0,-5.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[4.0,0.2,4.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_red",
+ "TRANS":[18.0,0.0,-8.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[3.0,3.0,3.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_blue",
+ "TRANS":[-15.0,3.0,-12.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.5,2.5,2.5]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[10.0,-8.0,15.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.8,2.8,2.8]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[-18.0,-2.0,5.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[3.5,3.5,3.5]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"mirror",
+ "TRANS":[12.0,10.0,10.0],
+ "ROTAT":[0.0,45.0,0.0],
+ "SCALE":[2.0,2.0,0.3]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[-8.0,-10.0,-15.0],
+ "ROTAT":[0.0,30.0,0.0],
+ "SCALE":[1.5,3.0,0.2]
+ }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/bvh_stress_scene.json b/scenes/bvh_stress_scene.json
new file mode 100644
index 000000000..7ec2cc8a7
--- /dev/null
+++ b/scenes/bvh_stress_scene.json
@@ -0,0 +1,245 @@
+{
+ "Materials":
+ {
+ "light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0, 1.0, 1.0],
+ "EMITTANCE":3.0,
+ "ROUGHNESS":1.0
+ },
+ "diffuse_white":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.98, 0.98, 0.98],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_red":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.85, 0.35, 0.35],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_green":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.35, 0.85, 0.35],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_blue":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.35, 0.35, 0.85],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_yellow":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.85, 0.85, 0.35],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_cyan":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.35, 0.85, 0.85],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_magenta":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.85, 0.35, 0.85],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_orange":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.95, 0.55, 0.25],
+ "ROUGHNESS":1.0
+ },
+ "specular_white":
+ {
+ "TYPE":"Specular",
+ "RGB":[0.98, 0.98, 0.98],
+ "ROUGHNESS":0.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[800,800],
+ "FOVY":45.0,
+ "ITERATIONS":5000,
+ "DEPTH":8,
+ "FILE":"cornell",
+ "EYE":[0.0,5.0,10.5],
+ "LOOKAT":[0.0,5.0,0.0],
+ "UP":[0.0,1.0,0.0]
+ },
+ "Objects":
+ [
+ {
+ "TYPE":"cube",
+ "MATERIAL":"light",
+ "TRANS":[0.0,14.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[8.0,0.3,8.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[0.0,0.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[15.0,0.01,15.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[0.0,14.0,0.0],
+ "ROTAT":[0.0,0.0,90.0],
+ "SCALE":[0.01,15.0,15.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_red",
+ "TRANS":[-7.5,7.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[0.01,14.0,15.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_green",
+ "TRANS":[7.5,7.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[0.01,14.0,15.0]
+ },
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[-6.0,1.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[-4.0,1.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[-2.0,1.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[0.0,1.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[2.0,1.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[4.0,1.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[6.0,1.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[-6.0,1.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[-4.0,1.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[-2.0,1.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[0.0,1.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[2.0,1.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[4.0,1.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[6.0,1.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[-6.0,1.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[-4.0,1.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[-2.0,1.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[0.0,1.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[2.0,1.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[4.0,1.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[6.0,1.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[-6.0,1.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[-4.0,1.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[-2.0,1.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[0.0,1.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[2.0,1.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[4.0,1.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[6.0,1.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[-6.0,1.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[-4.0,1.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[-2.0,1.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[0.0,1.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[2.0,1.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[4.0,1.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[6.0,1.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[-6.0,1.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[-4.0,1.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[-2.0,1.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[0.0,1.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[2.0,1.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[4.0,1.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[6.0,1.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[-6.0,3.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[-4.0,3.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[-2.0,3.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[0.0,3.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[2.0,3.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[4.0,3.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[6.0,3.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[-6.0,3.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[-4.0,3.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[-2.0,3.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[0.0,3.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[2.0,3.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[4.0,3.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[6.0,3.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[-6.0,3.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[-4.0,3.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[-2.0,3.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[0.0,3.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[2.0,3.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[4.0,3.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[6.0,3.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[-6.0,3.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[-4.0,3.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[-2.0,3.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[0.0,3.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[2.0,3.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[4.0,3.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[6.0,3.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[-6.0,3.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[-4.0,3.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[-2.0,3.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[0.0,3.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[2.0,3.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[4.0,3.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[6.0,3.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[-6.0,3.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[-4.0,3.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[-2.0,3.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[0.0,3.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[2.0,3.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[4.0,3.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[6.0,3.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[-6.0,5.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[-4.0,5.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[-2.0,5.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[0.0,5.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[2.0,5.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[4.0,5.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[6.0,5.0,-6.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[-6.0,5.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[-4.0,5.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[-2.0,5.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[0.0,5.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[2.0,5.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[4.0,5.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[6.0,5.0,-4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[-6.0,5.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[-4.0,5.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[-2.0,5.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[0.0,5.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[2.0,5.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[4.0,5.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[6.0,5.0,-2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[-6.0,5.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[-4.0,5.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[-2.0,5.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[0.0,5.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[2.0,5.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[4.0,5.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[6.0,5.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[-6.0,5.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[-4.0,5.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[-2.0,5.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[0.0,5.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[2.0,5.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[4.0,5.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[6.0,5.0,2.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_magenta","TRANS":[-6.0,5.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_orange","TRANS":[-4.0,5.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_red","TRANS":[-2.0,5.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_green","TRANS":[0.0,5.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_blue","TRANS":[2.0,5.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_yellow","TRANS":[4.0,5.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"sphere","MATERIAL":"diffuse_cyan","TRANS":[6.0,5.0,4.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.8,0.8,0.8]},
+ {"TYPE":"cube","MATERIAL":"specular_white","TRANS":[-3.0,0.75,-3.0],"ROTAT":[0.0,20.0,0.0],"SCALE":[1.2,1.5,1.2]},
+ {"TYPE":"cube","MATERIAL":"diffuse_orange","TRANS":[3.0,0.75,3.0],"ROTAT":[0.0,-25.0,0.0],"SCALE":[1.2,1.5,1.2]},
+ {"TYPE":"sphere","MATERIAL":"specular_white","TRANS":[-3.5,7.0,1.5],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.9,0.9,0.9]},
+ {"TYPE":"sphere","MATERIAL":"specular_white","TRANS":[3.5,7.5,-1.5],"ROTAT":[0.0,0.0,0.0],"SCALE":[0.9,0.9,0.9]}
+ ]
+}
\ No newline at end of file
diff --git a/scenes/bvh_test_scene.json b/scenes/bvh_test_scene.json
new file mode 100644
index 000000000..686eee169
--- /dev/null
+++ b/scenes/bvh_test_scene.json
@@ -0,0 +1,92 @@
+{
+ "Materials":
+ {
+ "light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0, 1.0, 1.0],
+ "EMITTANCE":5.0,
+ "ROUGHNESS":1.0
+ },
+ "diffuse_white":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.98, 0.98, 0.98],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_red":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.85, 0.35, 0.35],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_green":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.35, 0.85, 0.35],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_blue":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.35, 0.35, 0.85],
+ "ROUGHNESS":1.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[800,800],
+ "FOVY":45.0,
+ "ITERATIONS":5000,
+ "DEPTH":32,
+ "FILE":"bvh_test",
+ "EYE":[0.0,5.0,15.0],
+ "LOOKAT":[0.0,5.0,0.0],
+ "UP":[0.0,1.0,0.0]
+ },
+ "Objects":
+ [
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_red",
+ "TRANS":[-10.0,5.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.0,2.0,2.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_green",
+ "TRANS":[-6.0,5.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.0,2.0,2.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_blue",
+ "TRANS":[-2.0,5.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.0,2.0,2.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[2.0,5.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.0,2.0,2.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_red",
+ "TRANS":[6.0,5.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.0,2.0,2.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_green",
+ "TRANS":[10.0,5.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.0,2.0,2.0]
+ }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/bvh_test_scene_2.json b/scenes/bvh_test_scene_2.json
new file mode 100644
index 000000000..b9b2035c0
--- /dev/null
+++ b/scenes/bvh_test_scene_2.json
@@ -0,0 +1,119 @@
+{
+ "Materials":
+ {
+ "light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0, 1.0, 1.0],
+ "EMITTANCE":5.0,
+ "ROUGHNESS":1.0
+ },
+ "diffuse_white":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.98, 0.98, 0.98],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_red":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.85, 0.35, 0.35],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_green":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.35, 0.85, 0.35],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_blue":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.35, 0.35, 0.85],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_yellow":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.85, 0.85, 0.35],
+ "ROUGHNESS":1.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[800,800],
+ "FOVY":45.0,
+ "ITERATIONS":5000,
+ "DEPTH":32,
+ "FILE":"bvh_grid_test",
+ "EYE":[0.0,8.0,12.0],
+ "LOOKAT":[0.0,3.0,0.0],
+ "UP":[0.0,1.0,0.0]
+ },
+ "Objects":
+ [
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_red",
+ "TRANS":[-3.0,2.0,-3.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_green",
+ "TRANS":[0.0,2.0,-3.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_blue",
+ "TRANS":[3.0,2.0,-3.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_yellow",
+ "TRANS":[-3.0,2.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[0.0,2.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_red",
+ "TRANS":[3.0,2.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_green",
+ "TRANS":[-3.0,2.0,3.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_blue",
+ "TRANS":[0.0,2.0,3.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_yellow",
+ "TRANS":[3.0,2.0,3.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/bvh_test_scene_3.json b/scenes/bvh_test_scene_3.json
new file mode 100644
index 000000000..54cf5b6f0
--- /dev/null
+++ b/scenes/bvh_test_scene_3.json
@@ -0,0 +1,99 @@
+{
+ "Materials":
+ {
+ "mat_red":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[1.0, 0.0, 0.0],
+ "ROUGHNESS":1.0
+ },
+ "mat_green":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.0, 1.0, 0.0],
+ "ROUGHNESS":1.0
+ },
+ "mat_blue":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.0, 0.0, 1.0],
+ "ROUGHNESS":1.0
+ },
+ "mat_yellow":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[1.0, 1.0, 0.0],
+ "ROUGHNESS":1.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[800,800],
+ "FOVY":45.0,
+ "ITERATIONS":5000,
+ "DEPTH":8,
+ "FILE":"bvh_test",
+ "EYE":[0.0,0.0,20.0],
+ "LOOKAT":[0.0,0.0,0.0],
+ "UP":[0.0,1.0,0.0]
+ },
+ "Objects":
+ [
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mat_red",
+ "TRANS":[-4.0, 3.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.0, 1.0, 1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mat_green",
+ "TRANS":[-2.0, 3.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.0, 1.0, 1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mat_blue",
+ "TRANS":[2.0, 3.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.0, 1.0, 1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mat_yellow",
+ "TRANS":[4.0, 3.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.0, 1.0, 1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mat_red",
+ "TRANS":[-4.0, 6.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.0, 1.0, 1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mat_green",
+ "TRANS":[-2.0, 6.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.0, 1.0, 1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mat_blue",
+ "TRANS":[2.0, 6.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.0, 1.0, 1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mat_yellow",
+ "TRANS":[4.0, 6.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.0, 1.0, 1.0]
+ }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/bvh_test_scene_4.json b/scenes/bvh_test_scene_4.json
new file mode 100644
index 000000000..85c9d03c5
--- /dev/null
+++ b/scenes/bvh_test_scene_4.json
@@ -0,0 +1,368 @@
+{
+ "Materials":
+ {
+ "light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0, 1.0, 1.0],
+ "EMITTANCE":2.0,
+ "ROUGHNESS":1.0
+ },
+ "diffuse_white":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.98, 0.98, 0.98],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_red":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.85, 0.35, 0.35],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_green":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.35, 0.85, 0.35],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_blue":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.35, 0.35, 0.85],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_yellow":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.85, 0.85, 0.35],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_cyan":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.35, 0.85, 0.85],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_magenta":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.85, 0.35, 0.85],
+ "ROUGHNESS":1.0
+ },
+ "specular_white":
+ {
+ "TYPE":"Specular",
+ "RGB":[0.98, 0.98, 0.98],
+ "ROUGHNESS":0.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[800,800],
+ "FOVY":45.0,
+ "ITERATIONS":5000,
+ "DEPTH":24,
+ "FILE":"bvh_stress_test",
+ "EYE":[0.0,5.0,7.5],
+ "LOOKAT":[0.0,5.0,0.0],
+ "UP":[0.0,1.0,0.0]
+ },
+ "Objects":
+ [
+ {
+ "TYPE":"cube",
+ "MATERIAL":"light",
+ "TRANS":[0.0,10.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[4.0,0.3,4.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[0.0,0.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[10.0,0.01,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[0.0,10.0,0.0],
+ "ROTAT":[0.0,0.0,90.0],
+ "SCALE":[0.01,10.0,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_red",
+ "TRANS":[-5.0,5.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[0.01,10.0,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_green",
+ "TRANS":[5.0,5.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[0.01,10.0,10.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_red",
+ "TRANS":[-3.5,1.0,-3.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_green",
+ "TRANS":[-3.5,1.0,-1.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_blue",
+ "TRANS":[-3.5,1.0,0.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_yellow",
+ "TRANS":[-3.5,1.0,2.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_cyan",
+ "TRANS":[-1.5,1.0,-3.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_magenta",
+ "TRANS":[-1.5,1.0,-1.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[-1.5,1.0,0.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_red",
+ "TRANS":[-1.5,1.0,2.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_green",
+ "TRANS":[0.5,1.0,-3.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_blue",
+ "TRANS":[0.5,1.0,-1.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_yellow",
+ "TRANS":[0.5,1.0,0.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_cyan",
+ "TRANS":[0.5,1.0,2.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_magenta",
+ "TRANS":[2.5,1.0,-3.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[2.5,1.0,-1.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_red",
+ "TRANS":[2.5,1.0,0.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_green",
+ "TRANS":[2.5,1.0,2.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_blue",
+ "TRANS":[-3.5,3.0,-3.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_yellow",
+ "TRANS":[-3.5,3.0,-1.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_cyan",
+ "TRANS":[-3.5,3.0,0.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_magenta",
+ "TRANS":[-3.5,3.0,2.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[-1.5,3.0,-3.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_red",
+ "TRANS":[-1.5,3.0,-1.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_green",
+ "TRANS":[-1.5,3.0,0.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_blue",
+ "TRANS":[-1.5,3.0,2.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_yellow",
+ "TRANS":[0.5,3.0,-3.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_cyan",
+ "TRANS":[0.5,3.0,-1.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_magenta",
+ "TRANS":[0.5,3.0,0.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[0.5,3.0,2.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_red",
+ "TRANS":[2.5,3.0,-3.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_green",
+ "TRANS":[2.5,3.0,-1.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_blue",
+ "TRANS":[2.5,3.0,0.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_yellow",
+ "TRANS":[2.5,3.0,2.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"specular_white",
+ "TRANS":[-2.0,0.75,-2.0],
+ "ROTAT":[0.0,15.0,0.0],
+ "SCALE":[1.5,1.5,1.5]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_cyan",
+ "TRANS":[1.5,0.5,1.0],
+ "ROTAT":[0.0,30.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"specular_white",
+ "TRANS":[-2.5,5.0,1.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[0.8,0.8,0.8]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_magenta",
+ "TRANS":[1.8,5.5,-2.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[0.7,0.7,0.7]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[0.0,6.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[0.6,0.6,0.6]
+ }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/bvh_test_scene_5.json b/scenes/bvh_test_scene_5.json
new file mode 100644
index 000000000..6fd24079e
--- /dev/null
+++ b/scenes/bvh_test_scene_5.json
@@ -0,0 +1,85 @@
+{
+ "Materials":
+ {
+ "light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0, 1.0, 1.0],
+ "EMITTANCE":2.0,
+ "ROUGHNESS":1.0
+ },
+ "diffuse_white":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.98, 0.98, 0.98],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_grey":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.4, 0.4, 0.4],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_red":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.85, 0.35, 0.35],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_green":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.35, 0.85, 0.35],
+ "ROUGHNESS":1.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[800,800],
+ "FOVY":45.0,
+ "ITERATIONS":5000,
+ "DEPTH":24,
+ "FILE":"bvh_stress_test",
+ "EYE":[0.0,5.0,10.0],
+ "LOOKAT":[0.0,5.0,0.0],
+ "UP":[0.0,1.0,0.0]
+ },
+ "Objects":
+ [
+ {
+ "TYPE":"cube",
+ "MATERIAL":"light",
+ "TRANS":[0.0,10.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[4.0,0.3,4.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[0.0,0.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[10.0,0.01,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[0.0,10.0,0.0],
+ "ROTAT":[0.0,0.0,90.0],
+ "SCALE":[0.01,10.0,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_red",
+ "TRANS":[-5.0,5.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[0.01,10.0,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_green",
+ "TRANS":[5.0,5.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[0.01,10.0,10.0]
+ }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/cornell.json b/scenes/cornell.json
index e7419885e..ef9701a15 100644
--- a/scenes/cornell.json
+++ b/scenes/cornell.json
@@ -5,22 +5,26 @@
{
"TYPE":"Emitting",
"RGB":[1.0, 1.0, 1.0],
- "EMITTANCE":5.0
+ "EMITTANCE":5.0,
+ "ROUGHNESS":1.0
},
"diffuse_white":
{
"TYPE":"Diffuse",
- "RGB":[0.98, 0.98, 0.98]
+ "RGB":[0.98, 0.98, 0.98],
+ "ROUGHNESS":1.0
},
"diffuse_red":
{
"TYPE":"Diffuse",
- "RGB":[0.85, 0.35, 0.35]
+ "RGB":[0.85, 0.35, 0.35],
+ "ROUGHNESS":1.0
},
"diffuse_green":
{
"TYPE":"Diffuse",
- "RGB":[0.35, 0.85, 0.35]
+ "RGB":[0.35, 0.85, 0.35],
+ "ROUGHNESS":1.0
},
"specular_white":
{
@@ -36,7 +40,7 @@
"ITERATIONS":5000,
"DEPTH":8,
"FILE":"cornell",
- "EYE":[0.0,5.0,10.5],
+ "EYE":[0.0,5.0,7.5],
"LOOKAT":[0.0,5.0,0.0],
"UP":[0.0,1.0,0.0]
},
@@ -66,9 +70,9 @@
{
"TYPE":"cube",
"MATERIAL":"diffuse_white",
- "TRANS":[0.0,5.0,-5.0],
+ "TRANS":[0.30,5.0,-5.0],
"ROTAT":[0.0,90.0,0.0],
- "SCALE":[0.01,10.0,10.0]
+ "SCALE":[0.01,11.0,11.0]
},
{
"TYPE":"cube",
@@ -86,10 +90,59 @@
},
{
"TYPE":"sphere",
- "MATERIAL":"specular_white",
- "TRANS":[-1.0,4.0,-1.0],
- "ROTAT":[0.0,0.0,0.0],
- "SCALE":[3.0,3.0,3.0]
+ "MATERIAL":"mat_red",
+ "TRANS":[-4.0, 3.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.0, 1.0, 1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mat_green",
+ "TRANS":[-2.0, 3.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.0, 1.0, 1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mat_blue",
+ "TRANS":[2.0, 3.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.0, 1.0, 1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mat_yellow",
+ "TRANS":[4.0, 3.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.0, 1.0, 1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mat_red",
+ "TRANS":[-4.0, 6.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.0, 1.0, 1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mat_green",
+ "TRANS":[-2.0, 6.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.0, 1.0, 1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mat_blue",
+ "TRANS":[2.0, 6.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.0, 1.0, 1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mat_yellow",
+ "TRANS":[4.0, 6.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.0, 1.0, 1.0]
}
]
}
\ No newline at end of file
diff --git a/scenes/cornell_blackhole.json b/scenes/cornell_blackhole.json
new file mode 100644
index 000000000..4a01191b7
--- /dev/null
+++ b/scenes/cornell_blackhole.json
@@ -0,0 +1,102 @@
+{
+ "Materials":
+ {
+ "blackhole":
+ {
+ "TYPE":"BlackHole",
+ "RGB":[0.0,0.0,0.0],
+ "EMITTANCE":0.0,
+ "ROUGHNESS":1.0,
+ "INNERRAD": 2.0,
+ "OUTERRAD": 8.0
+ },
+ "light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0, 1.0, 1.0],
+ "EMITTANCE":5.0,
+ "ROUGHNESS":1.0
+ },
+ "diffuse_white":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.98, 0.98, 0.98],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_red":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.85, 0.35, 0.35],
+ "ROUGHNESS":1.0
+ },
+ "diffuse_green":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.35, 0.85, 0.35],
+ "ROUGHNESS":1.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[800,800],
+ "FOVY":45.0,
+ "ITERATIONS":5000,
+ "DEPTH":32,
+ "FILE":"cornell_blackhole",
+ "EYE":[0.0,5.0,10.5],
+ "LOOKAT":[0.0,5.0,0.0],
+ "UP":[0.0,1.0,0.0]
+ },
+ "Objects":
+ [
+ {
+ "TYPE":"cube",
+ "MATERIAL":"light",
+ "TRANS":[0.0,10.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[3.0,0.3,3.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[0.0,0.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[10.0,0.01,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[0.0,10.0,0.0],
+ "ROTAT":[0.0,0.0,90.0],
+ "SCALE":[0.01,10.0,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_white",
+ "TRANS":[0.0,5.0,-5.0],
+ "ROTAT":[0.0,90.0,0.0],
+ "SCALE":[0.01,10.0,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_red",
+ "TRANS":[-5.0,5.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[0.01,10.0,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"diffuse_green",
+ "TRANS":[5.0,5.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[0.01,10.0,10.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"blackhole",
+ "TRANS":[0.0,5.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[8.0,8.0,8.0]
+ }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/depth_of_field_test_scene.json b/scenes/depth_of_field_test_scene.json
new file mode 100644
index 000000000..985c89e1e
--- /dev/null
+++ b/scenes/depth_of_field_test_scene.json
@@ -0,0 +1,173 @@
+{
+ "Materials":
+ {
+ "area_light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0, 1.0, 1.0],
+ "EMITTANCE":15.0,
+ "ROUGHNESS":1.0
+ },
+ "red_diffuse":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.9, 0.2, 0.2],
+ "ROUGHNESS":1.0
+ },
+ "green_diffuse":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.2, 0.9, 0.3],
+ "ROUGHNESS":1.0
+ },
+ "blue_diffuse":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.2, 0.3, 0.9],
+ "ROUGHNESS":1.0
+ },
+ "yellow_diffuse":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.9, 0.9, 0.1],
+ "ROUGHNESS":1.0
+ },
+ "purple_diffuse":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.8, 0.2, 0.8],
+ "ROUGHNESS":1.0
+ },
+ "orange_diffuse":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[1.0, 0.5, 0.1],
+ "ROUGHNESS":1.0
+ },
+ "floor_white":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.8, 0.8, 0.8],
+ "ROUGHNESS":1.0
+ },
+ "wall_white":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.9, 0.9, 0.9],
+ "ROUGHNESS":1.0
+ },
+ "chrome_mirror":
+ {
+ "TYPE":"Specular",
+ "RGB":[0.9, 0.9, 0.9],
+ "ROUGHNESS":0.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[1024,1024],
+ "FOVY":45.0,
+ "ITERATIONS":2000,
+ "DEPTH":42,
+ "FILE":"dof_test",
+ "EYE":[0.0,2.0,12.0],
+ "LOOKAT":[0.0,2.0,0.0],
+ "UP":[0.0,1.0,0.0],
+ "APERTURE":0.15,
+ "FOCAL_DISTANCE":8.0
+ },
+ "Objects":
+ [
+ {
+ "TYPE":"cube",
+ "MATERIAL":"floor_white",
+ "TRANS":[0.0,0.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[20.0,0.1,25.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"wall_white",
+ "TRANS":[0.0,5.0,-12.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[20.0,10.0,0.1]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"area_light",
+ "TRANS":[0.0,8.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[6.0,0.1,6.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"red_diffuse",
+ "TRANS":[0.0,1.0,10.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"green_diffuse",
+ "TRANS":[-2.0,1.0,8.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"blue_diffuse",
+ "TRANS":[2.0,1.0,8.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"yellow_diffuse",
+ "TRANS":[-1.0,1.5,6.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.2,1.2,1.2]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"purple_diffuse",
+ "TRANS":[1.0,1.5,6.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.2,1.2,1.2]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"orange_diffuse",
+ "TRANS":[0.0,2.0,4.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.4,1.4,1.4]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"chrome_mirror",
+ "TRANS":[-3.0,1.5,2.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.5,1.5,1.5]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"chrome_mirror",
+ "TRANS":[3.0,1.5,2.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.5,1.5,1.5]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"red_diffuse",
+ "TRANS":[-4.0,1.0,0.0],
+ "ROTAT":[0.0,25.0,0.0],
+ "SCALE":[1.5,2.0,1.5]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"blue_diffuse",
+ "TRANS":[4.0,1.0,0.0],
+ "ROTAT":[0.0,-25.0,0.0],
+ "SCALE":[1.5,2.0,1.5]
+ }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/depth_of_field_test_scene_V2.json b/scenes/depth_of_field_test_scene_V2.json
new file mode 100644
index 000000000..88af67373
--- /dev/null
+++ b/scenes/depth_of_field_test_scene_V2.json
@@ -0,0 +1,80 @@
+{
+ "Materials":
+ {
+ "area_light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0, 1.0, 1.0],
+ "EMITTANCE":12.0,
+ "ROUGHNESS":1.0
+ },
+ "red_diffuse":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.9, 0.2, 0.2],
+ "ROUGHNESS":1.0
+ },
+ "green_diffuse":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.2, 0.9, 0.3],
+ "ROUGHNESS":1.0
+ },
+ "blue_diffuse":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.2, 0.3, 0.9],
+ "ROUGHNESS":1.0
+ },
+ "floor_white":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.8, 0.8, 0.8],
+ "ROUGHNESS":1.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[800, 600],
+ "FOVY":45.0,
+ "ITERATIONS":5000,
+ "DEPTH":8,
+ "FILE":"dof_comparison",
+ "EYE":[0.0, 2.5, 10.0],
+ "LOOKAT":[0.0, 2.0, 0.0],
+ "UP":[0.0, -3.0, 0.0],
+ "APERTURE":0.2,
+ "FOCAL_DISTANCE":6.0
+ },
+ "Objects":
+ [
+ {
+ "TYPE":"cube",
+ "MATERIAL":"floor_white",
+ "TRANS":[0.0, 0.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[15.0, 0.1, 15.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"area_light",
+ "TRANS":[0.0, 7.0, 0.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[5.0, 0.1, 5.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"red_diffuse",
+ "TRANS":[0.0, 1.5, 2.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.5, 1.5, 1.5]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"green_diffuse",
+ "TRANS":[0.0, 1.5, 6.0],
+ "ROTAT":[0.0, 0.0, 0.0],
+ "SCALE":[1.5, 1.5, 1.5]
+ }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/diskTest_v1.json b/scenes/diskTest_v1.json
new file mode 100644
index 000000000..8ac35fd2c
--- /dev/null
+++ b/scenes/diskTest_v1.json
@@ -0,0 +1,39 @@
+{
+ "Materials":
+ {
+ "diffuse_red":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.85, 0.35, 0.35],
+ "ROUGHNESS":1.0
+ },
+ "light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0,1.0,1.0],
+ "EMITTANCE":5.0,
+ "ROUGHNESS":1.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[800,800],
+ "FOVY":45.0,
+ "ITERATIONS":5000,
+ "DEPTH":8,
+ "FILE":"diskTest_v1",
+ "EYE":[0.0,5.0,10.5],
+ "LOOKAT":[0.0,5.0,0.0],
+ "UP":[0.0,1.0,0.0]
+ },
+ "Objects":
+ [
+ {
+ "TYPE":"disk",
+ "MATERIAL":"light",
+ "TRANS":[0.0,0.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[5.0,0.0,5.0]
+ }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/doubleBH_V1.json b/scenes/doubleBH_V1.json
new file mode 100644
index 000000000..44fb20356
--- /dev/null
+++ b/scenes/doubleBH_V1.json
@@ -0,0 +1,141 @@
+{
+ "Materials":
+ {
+ "blackhole":
+ {
+ "TYPE":"BlackHole",
+ "RGB":[0.6,0.6,1.0],
+ "EMITTANCE":10.0,
+ "ROUGHNESS":1.0,
+ "INNERRAD": 1.0,
+ "OUTERRAD": 12.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[800,800],
+ "FOVY":75.0,
+ "ITERATIONS":5000,
+ "DEPTH":8,
+ "FILE":"onetwentyeightBH",
+ "EYE":[0.0,0.0,130.0],
+ "LOOKAT":[0.0,0.0,0.0],
+ "UP":[0.0,1.0,0.0]
+ },
+ "Objects":
+ [
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-52.5,42.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-37.5,42.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-22.5,42.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-7.5,42.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[7.5,42.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[22.5,42.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[37.5,42.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[52.5,42.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-52.5,30.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-37.5,30.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-22.5,30.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-7.5,30.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[7.5,30.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[22.5,30.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[37.5,30.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[52.5,30.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-52.5,18.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-37.5,18.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-22.5,18.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-7.5,18.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[7.5,18.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[22.5,18.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[37.5,18.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[52.5,18.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-52.5,6.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-37.5,6.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-22.5,6.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-7.5,6.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[7.5,6.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[22.5,6.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[37.5,6.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[52.5,6.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-52.5,-6.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-37.5,-6.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-22.5,-6.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-7.5,-6.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[7.5,-6.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[22.5,-6.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[37.5,-6.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[52.5,-6.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-52.5,-18.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-37.5,-18.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-22.5,-18.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-7.5,-18.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[7.5,-18.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[22.5,-18.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[37.5,-18.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[52.5,-18.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-52.5,-30.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-37.5,-30.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-22.5,-30.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-7.5,-30.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[7.5,-30.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[22.5,-30.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[37.5,-30.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[52.5,-30.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-52.5,-42.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-37.5,-42.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-22.5,-42.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-7.5,-42.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[7.5,-42.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[22.5,-42.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[37.5,-42.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[52.5,-42.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-45.0,36.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-30.0,36.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-15.0,36.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[0.0,36.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[15.0,36.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[30.0,36.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[45.0,36.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-45.0,24.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-30.0,24.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-15.0,24.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[0.0,24.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[15.0,24.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[30.0,24.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[45.0,24.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-45.0,12.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-30.0,12.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-15.0,12.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[0.0,12.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[15.0,12.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[30.0,12.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[45.0,12.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-45.0,0.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-30.0,0.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-15.0,0.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[0.0,0.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[15.0,0.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[30.0,0.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[45.0,0.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-45.0,-12.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-30.0,-12.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-15.0,-12.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[0.0,-12.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[15.0,-12.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[30.0,-12.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[45.0,-12.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-45.0,-24.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-30.0,-24.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-15.0,-24.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[0.0,-24.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[15.0,-24.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[30.0,-24.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[45.0,-24.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-45.0,-36.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-30.0,-36.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[-15.0,-36.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[0.0,-36.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[15.0,-36.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[30.0,-36.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]},
+ {"TYPE":"sphere","MATERIAL":"blackhole","TRANS":[45.0,-36.0,0.0],"ROTAT":[0.0,0.0,0.0],"SCALE":[12.0,12.0,12.0]}
+ ]
+}
\ No newline at end of file
diff --git a/scenes/enhanced_scene.json b/scenes/enhanced_scene.json
new file mode 100644
index 000000000..97a6567ef
--- /dev/null
+++ b/scenes/enhanced_scene.json
@@ -0,0 +1,167 @@
+{
+ "Materials":
+ {
+ "warm_light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0, 0.85, 0.6],
+ "EMITTANCE":6.0,
+ "ROUGHNESS":1.0
+ },
+ "cool_light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[0.7, 0.9, 1.0],
+ "EMITTANCE":6.0,
+ "ROUGHNESS":1.0
+ },
+ "accent_light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0, 0.7, 0.9],
+ "EMITTANCE":6.5,
+ "ROUGHNESS":1.0
+ },
+ "wall_cream":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.95, 0.92, 0.85],
+ "ROUGHNESS":1.0
+ },
+ "wall_sage":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.7, 0.8, 0.7],
+ "ROUGHNESS":1.0
+ },
+ "wall_dusty_blue":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.6, 0.7, 0.85],
+ "ROUGHNESS":1.0
+ },
+ "high_reflective":
+ {
+ "TYPE":"Specular",
+ "RGB":[0.95, 0.95, 0.95],
+ "ROUGHNESS":0.0
+ },
+ "medium_reflective":
+ {
+ "TYPE":"Specular",
+ "RGB":[0.8, 0.85, 0.9],
+ "ROUGHNESS":0.0
+ },
+ "low_reflective":
+ {
+ "TYPE":"Specular",
+ "RGB":[0.6, 0.65, 0.7],
+ "ROUGHNESS":0.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[800,800],
+ "FOVY":45.0,
+ "ITERATIONS":5000,
+ "DEPTH":32,
+ "FILE":"cornell",
+ "EYE":[0.0,5.0,10.5],
+ "LOOKAT":[0.0,5.0,0.0],
+ "UP":[0.0,1.0,0.0]
+ },
+ "Objects":
+ [
+ {
+ "TYPE":"cube",
+ "MATERIAL":"wall_cream",
+ "TRANS":[0.0,0.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[10.0,0.01,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"wall_cream",
+ "TRANS":[0.0,10.0,0.0],
+ "ROTAT":[0.0,0.0,90.0],
+ "SCALE":[0.01,10.0,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"wall_cream",
+ "TRANS":[0.0,5.0,-5.0],
+ "ROTAT":[0.0,90.0,0.0],
+ "SCALE":[0.01,10.0,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"wall_sage",
+ "TRANS":[-5.0,5.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[0.01,10.0,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"wall_dusty_blue",
+ "TRANS":[5.0,5.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[0.01,10.0,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"warm_light",
+ "TRANS":[-2.5,7.5,1.5],
+ "ROTAT":[0.0,25.0,15.0],
+ "SCALE":[0.8,3.2,0.6]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"cool_light",
+ "TRANS":[3.2,2.8,-2.0],
+ "ROTAT":[45.0,60.0,0.0],
+ "SCALE":[2.8,0.4,2.8]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"accent_light",
+ "TRANS":[0.5,8.5,-1.0],
+ "ROTAT":[0.0,0.0,45.0],
+ "SCALE":[0.3,0.3,4.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"high_reflective",
+ "TRANS":[-3.2,1.8,-1.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[3.6,3.6,3.6]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"medium_reflective",
+ "TRANS":[1.8,1.0,3.2],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.0,2.0,2.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"low_reflective",
+ "TRANS":[2.5,0.6,-0.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.2,1.2,1.2]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"medium_reflective",
+ "TRANS":[-1.5,3.5,2.8],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.4,1.4,1.4]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"wall_sage",
+ "TRANS":[4.2,0.8,0.8],
+ "ROTAT":[0.0,30.0,0.0],
+ "SCALE":[0.8,1.6,0.8]
+ }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/mixed_scene.json b/scenes/mixed_scene.json
new file mode 100644
index 000000000..083d6a3ef
--- /dev/null
+++ b/scenes/mixed_scene.json
@@ -0,0 +1,153 @@
+{
+ "Materials":
+ {
+ "warm_light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0, 0.85, 0.6],
+ "EMITTANCE":5.5,
+ "ROUGHNESS":1.0
+ },
+ "cool_light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[0.7, 0.9, 1.0],
+ "EMITTANCE":5.0,
+ "ROUGHNESS":1.0
+ },
+ "accent_light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0, 0.7, 0.9],
+ "EMITTANCE":6.0,
+ "ROUGHNESS":1.0
+ },
+ "wall_cream":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.95, 0.92, 0.85],
+ "ROUGHNESS":1.0
+ },
+ "wall_sage":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.7, 0.8, 0.7],
+ "ROUGHNESS":1.0
+ },
+ "wall_dusty_blue":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.6, 0.7, 0.85],
+ "ROUGHNESS":1.0
+ },
+ "high_reflective":
+ {
+ "TYPE":"Specular",
+ "RGB":[0.95, 0.95, 0.95],
+ "ROUGHNESS":0.0
+ },
+ "medium_reflective":
+ {
+ "TYPE":"Specular",
+ "RGB":[0.8, 0.85, 0.9],
+ "ROUGHNESS":0.0
+ },
+ "low_reflective":
+ {
+ "TYPE":"Specular",
+ "RGB":[0.6, 0.65, 0.7],
+ "ROUGHNESS":0.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[800,800],
+ "FOVY":45.0,
+ "ITERATIONS":5000,
+ "DEPTH":32,
+ "FILE":"cornell",
+ "EYE":[0.0,5.0,10.5],
+ "LOOKAT":[0.0,5.0,0.0],
+ "UP":[0.0,1.0,0.0]
+ },
+ "Objects":
+ [
+ {
+ "TYPE":"cube",
+ "MATERIAL":"wall_cream",
+ "TRANS":[0.0,0.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[10.0,0.01,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"wall_cream",
+ "TRANS":[0.0,10.0,0.0],
+ "ROTAT":[0.0,0.0,90.0],
+ "SCALE":[0.01,10.0,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"wall_cream",
+ "TRANS":[0.0,5.0,-5.0],
+ "ROTAT":[0.0,90.0,0.0],
+ "SCALE":[0.01,10.0,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"wall_sage",
+ "TRANS":[-5.0,5.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[0.01,10.0,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"wall_dusty_blue",
+ "TRANS":[5.0,5.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[0.01,10.0,10.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"warm_light",
+ "TRANS":[-2.0,2.0,2.0],
+ "ROTAT":[0.0,15.0,0.0],
+ "SCALE":[1.5,1.5,1.5]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"cool_light",
+ "TRANS":[3.0,1.5,-1.5],
+ "ROTAT":[0.0,30.0,0.0],
+ "SCALE":[1.2,1.2,1.2]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"accent_light",
+ "TRANS":[0.0,3.5,0.0],
+ "ROTAT":[0.0,45.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"high_reflective",
+ "TRANS":[-3.0,2.5,-2.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.0,2.0,2.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"medium_reflective",
+ "TRANS":[2.5,1.8,2.5],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.6,1.6,1.6]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"low_reflective",
+ "TRANS":[0.5,1.2,-3.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.2,1.2,1.2]
+ }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/obj_test_scene.json b/scenes/obj_test_scene.json
new file mode 100644
index 000000000..460d510b6
--- /dev/null
+++ b/scenes/obj_test_scene.json
@@ -0,0 +1,46 @@
+{
+ "Materials": {
+ "light_white": {
+ "TYPE": "Emitting",
+ "RGB": [1.0, 1.0, 1.0],
+ "EMITTANCE": 15.0,
+ "ROUGHNESS": 1.0
+ },
+ "mesh_diffuse": {
+ "TYPE": "Diffuse",
+ "RGB": [0.8, 0.8, 0.8],
+ "ROUGHNESS": 1.0
+ },
+ "mesh_mirror": {
+ "TYPE": "Specular",
+ "RGB": [0.95, 0.95, 0.95],
+ "ROUGHNESS": 0.0
+ },
+ "mesh_colored": {
+ "TYPE": "Diffuse",
+ "RGB": [0.5, 0.5, 0.5],
+ "ROUGHNESS": 1.0
+ }
+ },
+
+ "Camera": {
+ "RES": [800, 800],
+ "FOVY": 45.0,
+ "ITERATIONS": 5000,
+ "DEPTH": 8,
+ "FILE": "obj_test",
+ "EYE": [0.0, 3.0, 10.0],
+ "LOOKAT": [0.0, 0.0, 0.0],
+ "UP": [0.0, 1.0, 0.0]
+ },
+
+ "Objects": [
+ {
+ "TYPE": "sphere",
+ "MATERIAL": "light_white",
+ "TRANS": [0.0, 10.0, 0.0],
+ "ROTAT": [0.0, 0.0, 0.0],
+ "SCALE": [1.0, 1.0, 1.0]
+ }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/singleBH_V1.json b/scenes/singleBH_V1.json
new file mode 100644
index 000000000..2ea6a89d3
--- /dev/null
+++ b/scenes/singleBH_V1.json
@@ -0,0 +1,46 @@
+{
+ "Materials":
+ {
+ "blackhole":
+ {
+ "TYPE":"BlackHole",
+ "RGB": [0.8, 0.3, 0.9],
+ "EMITTANCE":10.0,
+ "ROUGHNESS":1.0,
+ "INNERRAD": 1.0,
+ "OUTERRAD": 16.0
+ },
+ "mirror_chrome": {
+ "TYPE": "Specular",
+ "RGB": [0.98, 0.98, 0.98],
+ "ROUGHNESS": 0.0
+ },
+ "light_blue": {
+ "TYPE": "Emitting",
+ "RGB": [0.5, 0.7, 1.0],
+ "EMITTANCE": 12.0,
+ "ROUGHNESS": 1.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[1080, 800],
+ "FOVY":30.0,
+ "ITERATIONS":5000,
+ "DEPTH":8,
+ "FILE":"singleBH_V1",
+ "EYE":[0.0,-5.0,18.0],
+ "LOOKAT":[0.0,0.0,0.0],
+ "UP":[0.0,1.0,0.0]
+ },
+ "Objects":
+ [
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"blackhole",
+ "TRANS":[0.0,0.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[16.0,16.0,16.0]
+ }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/singleBH_V1_mirrorBox.json b/scenes/singleBH_V1_mirrorBox.json
new file mode 100644
index 000000000..fe2ea5132
--- /dev/null
+++ b/scenes/singleBH_V1_mirrorBox.json
@@ -0,0 +1,49 @@
+{
+ "Materials":
+ {
+ "blackhole":
+ {
+ "TYPE":"BlackHole",
+ "RGB":[0.0,0.0,0.0],
+ "EMITTANCE":0.0,
+ "ROUGHNESS":1.0,
+ "INNERRAD": 1.0,
+ "OUTERRAD": 12.0
+ },
+ "mirror":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[1.0, 1.0, 1.0],
+ "EMITTANCE":0.0,
+ "ROUGHNESS":0.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[800,800],
+ "FOVY":30.0,
+ "ITERATIONS":5000,
+ "DEPTH":8,
+ "FILE":"mirrorbox_bh",
+ "EYE":[0.0,5.0,15.0],
+ "LOOKAT":[0.0,0.0,0.0],
+ "UP":[0.0,1.0,0.0]
+ },
+ "Objects":
+ [
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"blackhole",
+ "TRANS":[0.0,0.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[12.0,12.0,12.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"mirror",
+ "TRANS":[0.0,0.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[50.0, 50.0, 50.0]
+ }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/space_scene_bh.json b/scenes/space_scene_bh.json
new file mode 100644
index 000000000..f2b79e472
--- /dev/null
+++ b/scenes/space_scene_bh.json
@@ -0,0 +1,212 @@
+{
+ "Materials":
+ {
+ "blackhole":
+ {
+ "TYPE":"BlackHole",
+ "RGB":[1.0, 0.6, 0.2],
+ "EMITTANCE":5.0,
+ "ROUGHNESS":1.0,
+ "INNERRAD": 1.5,
+ "OUTERRAD": 20.0
+ },
+ "orange_light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0, 0.4, 0.1],
+ "EMITTANCE":25.0,
+ "ROUGHNESS":1.0
+ },
+ "blue_light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[0.2, 0.5, 1.0],
+ "EMITTANCE":20.0,
+ "ROUGHNESS":1.0
+ },
+ "cyan_light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[0.1, 0.8, 0.9],
+ "EMITTANCE":15.0,
+ "ROUGHNESS":1.0
+ },
+ "mirror":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.95, 0.95, 0.98],
+ "EMITTANCE":0.0,
+ "ROUGHNESS":0.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[1920,1080],
+ "FOVY":35.0,
+ "ITERATIONS":5000,
+ "DEPTH":12,
+ "FILE":"space_scene_bh",
+ "EYE":[8.0,18.0,12.0],
+ "LOOKAT":[0.0,0.0,0.0],
+ "UP":[0.0,1.0,0.0],
+ "LENSRADIUS": 0.4,
+ "FOCALDIST": 22.8
+ },
+ "Objects":
+ [
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"blackhole",
+ "TRANS":[0.0,0.0,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[20.0,20.0,20.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"orange_light",
+ "TRANS":[-28.0,8.0,-15.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.5,1.5,1.5]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"blue_light",
+ "TRANS":[25.0,-8.0,20.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.8,1.8,1.8]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"cyan_light",
+ "TRANS":[18.0,22.0,10.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.2,1.2,1.2]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[-24.0,-4.0,22.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[5.5,5.5,5.5]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[26.0,6.0,18.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.8,1.8,1.8]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[-20.0,22.0,-18.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[4.2,4.2,4.2]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[28.0,3.0,-26.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.5,2.5,2.5]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[-22.0,-18.0,14.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[3.6,3.6,3.6]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[10.0,16.0,-28.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.5,1.5,1.5]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[-12.0,-24.0,-14.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[4.8,4.8,4.8]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[24.0,12.0,26.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.2,2.2,2.2]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[30.0,-10.0,-6.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[3.0,3.0,3.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[-16.0,18.0,30.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.2,1.2,1.2]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[8.0,-26.0,16.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[3.8,3.8,3.8]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[-28.0,-12.0,-22.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.8,2.8,2.8]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[22.0,26.0,-14.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[4.5,4.5,4.5]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[-8.0,-6.0,32.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.6,1.6,1.6]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[26.0,-22.0,12.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[5.0,5.0,5.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[-26.0,14.0,-24.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.0,2.0,2.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[16.0,24.0,20.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[3.4,3.4,3.4]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"mirror",
+ "TRANS":[-30.0,-8.0,16.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.4,1.4,1.4]
+ }
+ ]
+}
\ No newline at end of file
diff --git a/scenes/sphere.json b/scenes/sphere.json
index affc3d750..a28839f91 100644
--- a/scenes/sphere.json
+++ b/scenes/sphere.json
@@ -5,7 +5,13 @@
{
"TYPE":"Emitting",
"RGB":[1.0,1.0,1.0],
- "EMITTANCE":5.0
+ "EMITTANCE":5.0,
+ "ROUGHNESS":1.0
+ },
+ "mirror_chrome": {
+ "TYPE": "Specular",
+ "RGB": [0.98, 0.98, 0.98],
+ "ROUGHNESS": 0.0
}
},
"Camera":
diff --git a/scenes/stream_compaction_test_scene.json b/scenes/stream_compaction_test_scene.json
new file mode 100644
index 000000000..6377c7c1b
--- /dev/null
+++ b/scenes/stream_compaction_test_scene.json
@@ -0,0 +1,155 @@
+{
+ "Materials":
+ {
+ "bright_light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0, 1.0, 1.0],
+ "EMITTANCE":25.0,
+ "ROUGHNESS":1.0
+ },
+ "warm_spot":
+ {
+ "TYPE":"Emitting",
+ "RGB":[1.0, 0.7, 0.4],
+ "EMITTANCE":15.0,
+ "ROUGHNESS":1.0
+ },
+ "cool_spot":
+ {
+ "TYPE":"Emitting",
+ "RGB":[0.4, 0.7, 1.0],
+ "EMITTANCE":12.0,
+ "ROUGHNESS":1.0
+ },
+ "distant_light":
+ {
+ "TYPE":"Emitting",
+ "RGB":[0.9, 0.9, 1.0],
+ "EMITTANCE":8.0,
+ "ROUGHNESS":1.0
+ },
+ "floor_gray":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.8, 0.8, 0.8],
+ "ROUGHNESS":1.0
+ },
+ "wall_light":
+ {
+ "TYPE":"Diffuse",
+ "RGB":[0.9, 0.9, 0.9],
+ "ROUGHNESS":1.0
+ },
+ "high_mirror":
+ {
+ "TYPE":"Specular",
+ "RGB":[0.95, 0.95, 0.95],
+ "ROUGHNESS":0.0
+ },
+ "partial_mirror":
+ {
+ "TYPE":"Specular",
+ "RGB":[0.7, 0.8, 0.9],
+ "ROUGHNESS":0.0
+ }
+ },
+ "Camera":
+ {
+ "RES":[800,800],
+ "FOVY":45.0,
+ "ITERATIONS":8000,
+ "DEPTH":8,
+ "FILE":"compaction_test",
+ "EYE":[0.0,3.0,15.0],
+ "LOOKAT":[0.0,2.0,0.0],
+ "UP":[0.0,1.0,0.0]
+ },
+ "Objects":
+ [
+ {
+ "TYPE":"cube",
+ "MATERIAL":"floor_gray",
+ "TRANS":[0.0,-0.1,0.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[50.0,0.2,50.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"wall_light",
+ "TRANS":[0.0,10.0,-25.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[50.0,20.0,0.2]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"bright_light",
+ "TRANS":[0.0,8.0,2.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[3.0,0.1,3.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"warm_spot",
+ "TRANS":[-8.0,4.0,-5.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.5,1.5,1.5]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"cool_spot",
+ "TRANS":[8.0,5.0,-3.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.0,2.0,2.0]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"distant_light",
+ "TRANS":[0.0,2.0,-20.0],
+ "ROTAT":[0.0,45.0,0.0],
+ "SCALE":[1.0,4.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"high_mirror",
+ "TRANS":[-3.0,1.5,8.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[3.0,3.0,3.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"partial_mirror",
+ "TRANS":[4.0,2.0,5.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[2.5,2.5,2.5]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"partial_mirror",
+ "TRANS":[-6.0,1.0,0.0],
+ "ROTAT":[0.0,30.0,0.0],
+ "SCALE":[1.0,2.0,0.2]
+ },
+ {
+ "TYPE":"cube",
+ "MATERIAL":"high_mirror",
+ "TRANS":[6.0,0.8,3.0],
+ "ROTAT":[0.0,-20.0,0.0],
+ "SCALE":[0.8,1.6,0.1]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"distant_light",
+ "TRANS":[12.0,6.0,-10.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.0,1.0,1.0]
+ },
+ {
+ "TYPE":"sphere",
+ "MATERIAL":"warm_spot",
+ "TRANS":[-12.0,3.0,-8.0],
+ "ROTAT":[0.0,0.0,0.0],
+ "SCALE":[1.2,1.2,1.2]
+ }
+ ]
+}
\ No newline at end of file
diff --git a/src/interactions.cu b/src/interactions.cu
index 676d8f814..f966edfde 100644
--- a/src/interactions.cu
+++ b/src/interactions.cu
@@ -44,6 +44,46 @@ __host__ __device__ glm::vec3 calculateRandomDirectionInHemisphere(
+ sin(around) * over * perpendicularDirection2;
}
+__host__ __device__ glm::vec3 cosWeightedHemiSphereSample(glm::vec3 normal,
+ thrust::default_random_engine& rng) {
+ thrust::uniform_real_distribution u01(-1, 1);
+ glm::vec2 sample = glm::vec2(u01(rng), u01(rng));
+ float pi = 3.14159265359;
+
+ // MALLY"S METHOD WITH DISK SAMPLE
+ float x = 2.0f * sample[0] - 1.0f;
+ float y = 2.0f * sample[1] - 1.0f;
+ float r = 1.f;
+ float a = 0.f;
+ if (x == 0 && y == 0) {
+ return normal;
+ }
+ if (abs(x) > abs(y)) {
+ r *= x;
+ a = (pi * 0.25) * (y / x);
+ }
+ else {
+ r *= y;
+ a = (pi * 0.5) - ((pi * 0.25) * (x / y));
+ }
+
+
+ glm::vec3 localSample = glm::vec3(cos(a) * r, sin(a) * r, sqrt(1 - (r * r)));
+
+ glm::vec3 nt = (abs(normal.x) > 0.1f) ? glm::vec3(0, 1, 0) : glm::vec3(1, 0, 0);
+ glm::vec3 tangent = glm::normalize(glm::cross(nt, normal));
+ glm::vec3 bitangent = glm::cross(normal, tangent);
+
+ return localSample.x * tangent + localSample.y * bitangent + localSample.z * normal;
+}
+
+__host__ __device__ glm::vec3 reflect(glm::vec3 normal, glm::vec3 wi) {
+ glm::vec3 cross = glm::normalize(glm::cross(normal, wi));
+ glm::vec3 planeNorm = glm::normalize(glm::cross(cross, normal));
+ return wi - 2.0f * glm::dot(wi, planeNorm) * planeNorm;
+
+}
+
__host__ __device__ void scatterRay(
PathSegment & pathSegment,
glm::vec3 intersect,
@@ -53,5 +93,186 @@ __host__ __device__ void scatterRay(
{
// TODO: implement this.
// A basic implementation of pure-diffuse shading will just call the
- // calculateRandomDirectionInHemisphere defined above.
+ // calculateRandomDirectionInHemisphere defined above
+
+ // CHANGE LATER TO BE BASED OFF OF MATERIAL TYPE
+
+ Ray newRay = Ray();
+ thrust::uniform_real_distribution u01(0, 1);
+ glm::vec3 newRayDir;
+ if (u01(rng) > 1.f - m.specular.exponent) {
+ //DIFFUSE
+ newRayDir = calculateRandomDirectionInHemisphere(normal, rng);
+ }
+ else {
+ //MIRROR
+ newRayDir = reflect(normal, -glm::normalize(pathSegment.ray.direction));
+
+ }
+ newRay.direction = glm::normalize(newRayDir);
+ newRay.origin = intersect + normal * 0.001f;
+ pathSegment.ray = newRay;
+ pathSegment.remainingBounces--;
+}
+
+__host__ __device__
+float windowWeight(float r, float oRad, float iRad) {
+ float normalizedDist = (r - iRad) / (oRad - iRad);
+ normalizedDist = glm::clamp(normalizedDist, 0.0f, 1.0f);
+ float shapedFalloff = glm::pow(1.0f - normalizedDist, 1.0f) * glm::smoothstep(0.0f, 0.1f, normalizedDist);
+ return glm::abs(1.0f - normalizedDist);
+}
+
+__host__ __device__ glm::vec3 bhAccel(glm::vec3 r, float h2, float M, float w) {
+ float rL = glm::length(r);
+ float radL5 = rL * rL * rL * rL * rL;
+ return (-3.0f * M * h2) * (1.0f / radL5) * r * w;
+}
+
+__host__ __device__ void rk4Step(glm::vec3& r, glm::vec3& v, float h2, float dt, float M, float w) {
+ // K1
+ glm::vec3 K1r = v;
+ glm::vec3 K1v = bhAccel(r, h2, M, w);
+
+ // K2
+ glm::vec3 K2r = v + 0.5f * dt * K1v;
+ glm::vec3 K2v = bhAccel(r + 0.5f * dt * K1r, h2, M, w);
+
+ // K3
+ glm::vec3 K3r = v + 0.5f * dt * K2v;
+ glm::vec3 K3v = bhAccel(r + 0.5f * dt * K2r, h2, M, w);
+
+ // K4
+ glm::vec3 K4r = v + dt * K3v;
+ glm::vec3 K4v = bhAccel(r + dt * K3r, h2, M, w);
+
+ r += (dt / 6.0f) * (K1r + 2.0f * K2r + 2.0f * K3r + K4r);
+ v += (dt / 6.0f) * (K1v + 2.0f * K2v + 2.0f * K3v + K4v);
+}
+
+__host__ __device__ float chooseDt(glm::vec3 r, glm::vec3 v, float h2, float M, float w, float iRad, float oRad) {
+ glm::vec3 a = bhAccel(r, h2, M, w);
+ glm::vec3 vPerp = glm::normalize(v);
+ glm::vec3 aPerp = a - glm::dot(a, vPerp) * vPerp;
+ float dtMin = 0.001f * iRad;
+ float dtMax = 0.1f * oRad;
+ float eps = 0.2f;
+ float amin = 1e-6f;
+ float dt = (eps * glm::length(v)) / glm::max(glm::length(aPerp), amin);
+ return glm::clamp(dt, dtMin, dtMax);
+}
+
+
+// 2D hash function for pseudorandom gradients
+__host__ __device__ inline glm::vec2 random2(glm::vec2 p) {
+ p = glm::vec2(glm::dot(p, glm::vec2(127.1f, 311.7f)),
+ glm::dot(p, glm::vec2(269.5f, 183.3f)));
+ return glm::fract(glm::sin(p) * 43758.5453123f) * 2.0f - 1.0f;
+}
+
+// 2D Perlin noise
+__host__ __device__ float noise(glm::vec2 x) {
+ // grid
+ glm::vec2 p = glm::floor(x);
+ glm::vec2 w = glm::fract(x);
+
+ // quintic interpolant
+ glm::vec2 u = w * w * w * (w * (w * 6.0f - 15.0f) + 10.0f);
+
+ // gradients
+ glm::vec2 ga = random2(p + glm::vec2(0.0f, 0.0f));
+ glm::vec2 gb = random2(p + glm::vec2(1.0f, 0.0f));
+ glm::vec2 gc = random2(p + glm::vec2(0.0f, 1.0f));
+ glm::vec2 gd = random2(p + glm::vec2(1.0f, 1.0f));
+
+ // projections
+ float va = glm::dot(ga, w - glm::vec2(0.0f, 0.0f));
+ float vb = glm::dot(gb, w - glm::vec2(1.0f, 0.0f));
+ float vc = glm::dot(gc, w - glm::vec2(0.0f, 1.0f));
+ float vd = glm::dot(gd, w - glm::vec2(1.0f, 1.0f));
+
+ // interpolation
+ return va +
+ u.x * (vb - va) +
+ u.y * (vc - va) +
+ u.x * u.y * (va - vb - vc + vd);
+}
+
+__host__ __device__ glm::vec2 swirl(glm::vec2 p, float swirlFactor) {
+ float r = glm::length(p);
+ float theta = glm::atan(p.y, p.x);
+ theta += swirlFactor * r;
+ return glm::vec2(r * glm::cos(theta), r * glm::sin(theta));
+}
+
+__host__ __device__ void blackHoleRay(
+ PathSegment& pathSegment,
+ glm::vec3 intersect,
+ glm::vec3 normal,
+ glm::mat4 invTrans,
+ const Material& m,
+ thrust::default_random_engine& rng
+) {
+ glm::vec3 bhCenter = intersect - normal * m.blackHole.oRad;
+ glm::vec3 r = intersect - bhCenter;
+ glm::vec3 prevR = r;
+ glm::vec3 v = glm::normalize(pathSegment.ray.direction);
+ float h2 = glm::length(glm::cross(r, v)) * glm::length(glm::cross(r, v));
+ float M = 0.5 * m.blackHole.iRad;
+ float w = 1.0f;
+ int maxSteps = 1024;
+ float dt = 0.001f;
+
+ glm::vec3 diskNorm = glm::vec3((invTrans * glm::vec4(0.0, 1.0, 0.0, 0.0)));
+
+ Ray newRay = Ray();
+ for (int i = 0; i < maxSteps; i++) {
+ if (prevR.y * r.y < 0.0) {
+ float t = prevR.y / (prevR.y - r.y);
+ glm::vec3 crossingPoint = prevR + t * (r - prevR);
+
+ float distFromCenter = glm::length(crossingPoint);
+ float normalizedDist = (distFromCenter - m.blackHole.iRad) / (m.blackHole.oRad - m.blackHole.iRad);
+ normalizedDist = glm::clamp(normalizedDist, 0.0f, 1.0f);
+
+ glm::vec2 swirlPos = swirl(glm::vec2(crossingPoint.x, crossingPoint.z), 0.4);
+
+ float shapedFalloff = glm::pow(1.0f - normalizedDist, 1.0f) * glm::smoothstep(0.0f, 0.1f, normalizedDist);
+ shapedFalloff = shapedFalloff * 0.6f + shapedFalloff * noise(swirlPos);
+
+ thrust::uniform_real_distribution u01(0, 1);
+ if (u01(rng) < shapedFalloff) {
+ pathSegment.color *= m.color * shapedFalloff * m.emittance;
+ pathSegment.remainingBounces = -1;
+ return;
+ }
+ }
+ if (glm::length(r) < m.blackHole.iRad) {
+ pathSegment.color *= glm::vec3(0.0);
+ pathSegment.remainingBounces = -1;
+ return;
+ }
+ if (glm::length(r) > m.blackHole.oRad && glm::dot(r, v) > 0.0f) {
+ newRay.direction = glm::normalize(v);
+ newRay.origin = bhCenter + r;
+ newRay.origin += newRay.direction * 0.001f;
+ pathSegment.ray = newRay;
+ pathSegment.remainingBounces--;
+ return;
+ }
+
+
+ w = windowWeight(glm::length(r), m.blackHole.oRad, m.blackHole.iRad);
+ prevR = r;
+ // RH4 STEP
+ dt = chooseDt(r, v, h2, M, w, m.blackHole.iRad, m.blackHole.oRad);
+ rk4Step(r, v, h2, dt, M, w);
+ }
+
+ newRay.direction = glm::normalize(v);
+ newRay.origin = bhCenter + r;
+ newRay.origin += newRay.direction * 0.001f;
+ pathSegment.ray = newRay;
+ pathSegment.remainingBounces--;
+ return;
}
diff --git a/src/interactions.h b/src/interactions.h
index a8b5f77ab..d29f5f3e0 100644
--- a/src/interactions.h
+++ b/src/interactions.h
@@ -46,3 +46,11 @@ __host__ __device__ void scatterRay(
glm::vec3 normal,
const Material& m,
thrust::default_random_engine& rng);
+
+__host__ __device__ void blackHoleRay(
+ PathSegment& pathSegment,
+ glm::vec3 intersect,
+ glm::vec3 normal,
+ glm::mat4 invTrans,
+ const Material& m,
+ thrust::default_random_engine& rng);
diff --git a/src/intersections.cu b/src/intersections.cu
index 0f9cd3d15..dc7216ee5 100644
--- a/src/intersections.cu
+++ b/src/intersections.cu
@@ -56,6 +56,7 @@ __host__ __device__ float boxIntersectionTest(
return -1;
}
+
__host__ __device__ float sphereIntersectionTest(
Geom sphere,
Ray r,
@@ -111,3 +112,73 @@ __host__ __device__ float sphereIntersectionTest(
return glm::length(r.origin - intersectionPoint);
}
+
+__host__ __device__ float diskIntersectionTest(
+ Geom disk,
+ Ray r,
+ glm::vec3& intersectionPoint,
+ glm::vec3& normal)
+{
+ glm::vec3 diskCenter = disk.translation;
+ glm::mat4 rotateY = glm::rotate(glm::mat4(), disk.rotation.y, glm::vec3(0.0, 1.0, 0.0));
+ glm::mat4 rotateX = glm::rotate(rotateY, disk.rotation.x, glm::vec3(1.0, 0.0, 0.0));
+ glm::mat4 rotateZ = glm::rotate(rotateX, disk.rotation.z, glm::vec3(0.0, 0.0, 1.0));
+ glm::vec3 diskNorm = glm::vec3(rotateZ * glm::vec4(0.0, 1.0, 0.0, 0.0));
+
+ float t = 0.0;
+
+ float denom = glm::dot(diskNorm, r.direction);
+ if (abs(denom) > 1e-6) {
+ glm::vec3 p0l0 = diskCenter - r.origin;
+ t = glm::dot(p0l0, diskNorm) / denom;
+ }
+
+ if (t > 0.0) {
+ glm::vec3 p = r.origin + r.direction * t;
+ glm::vec3 v = p - diskCenter;
+ float d2 = glm::dot(v, v);
+
+ if (d2 > disk.scale.x * disk.scale.x) {
+ t = 0.0;
+ }
+ }
+
+ if (t > 0.0) {
+ intersectionPoint = getPointOnRay(r, t);
+ normal = diskNorm;
+ }
+
+ return t;
+}
+
+__host__ __device__ float triangleIntersectionTest(
+ Geom triangle,
+ Ray r,
+ glm::vec3& intersectionPoint,
+ glm::vec3& normal,
+ bool& outside)
+{
+ glm::vec3 hitPos;
+ float t;
+
+ if (glm::intersectRayTriangle(r.origin, r.direction,
+ triangle.triPos1, triangle.triPos2, triangle.triPos3,
+ hitPos))
+ {
+
+ t = glm::length(hitPos - r.origin);
+ if (t > 0.0f) {
+ intersectionPoint = hitPos;
+
+ normal = glm::normalize(glm::cross(
+ triangle.triPos2 - triangle.triPos1,
+ triangle.triPos3 - triangle.triPos1
+ ));
+
+ outside = glm::dot(r.direction, normal) < 0.0f;
+ return t;
+ }
+ }
+
+ return -1.0f;
+}
diff --git a/src/intersections.h b/src/intersections.h
index 1f6e5a545..94575eb5a 100644
--- a/src/intersections.h
+++ b/src/intersections.h
@@ -3,6 +3,7 @@
#include "sceneStructs.h"
#include
+#include
#include
@@ -71,3 +72,16 @@ __host__ __device__ float sphereIntersectionTest(
glm::vec3& intersectionPoint,
glm::vec3& normal,
bool& outside);
+
+__host__ __device__ float diskIntersectionTest(
+ Geom disk,
+ Ray r,
+ glm::vec3& intersectionPoint,
+ glm::vec3& normal);
+
+__host__ __device__ float triangleIntersectionTest(
+ Geom triangle,
+ Ray r,
+ glm::vec3& intersectionPoint,
+ glm::vec3& normal,
+ bool& outside);
\ No newline at end of file
diff --git a/src/main.cpp b/src/main.cpp
index 330725251..e109f45bf 100644
--- a/src/main.cpp
+++ b/src/main.cpp
@@ -64,6 +64,7 @@ void runCuda();
void keyCallback(GLFWwindow *window, int key, int scancode, int action, int mods);
void mousePositionCallback(GLFWwindow* window, double xpos, double ypos);
void mouseButtonCallback(GLFWwindow* window, int button, int action, int mods);
+void scrollCallback(GLFWwindow* window, double xoffset, double yoffset);
std::string currentTimeString()
{
@@ -219,6 +220,7 @@ bool init()
glfwSetKeyCallback(window, keyCallback);
glfwSetCursorPosCallback(window, mousePositionCallback);
glfwSetMouseButtonCallback(window, mouseButtonCallback);
+ glfwSetScrollCallback(window, scrollCallback);
// Set up GL context
glewExperimental = GL_TRUE;
@@ -254,6 +256,10 @@ void InitImguiData(GuiDataContainer* guiData)
imguiData = guiData;
}
+static std::vector fpsHistory;
+static float fpsAccumulationTime = 0.0f;
+static float averageFPS = 0.0f;
+static const float FPS_SAMPLE_DURATION = 30.0f; // 30 seconds
// LOOK: Un-Comment to check ImGui Usage
void RenderImGui()
@@ -284,6 +290,29 @@ void RenderImGui()
// counter++;
//ImGui::SameLine();
//ImGui::Text("counter = %d", counter);
+
+ float currentFPS = ImGui::GetIO().Framerate;
+ float deltaTime = ImGui::GetIO().DeltaTime;
+
+ fpsHistory.push_back(currentFPS);
+ fpsAccumulationTime += deltaTime;
+
+ // Calculate average every 30 seconds
+ if (fpsAccumulationTime >= FPS_SAMPLE_DURATION) {
+ float sum = 0.0f;
+ for (float fps : fpsHistory) {
+ sum += fps;
+ }
+ averageFPS = sum / fpsHistory.size();
+
+ // Print to console
+ printf("30-second Average FPS: %.2f\n", averageFPS);
+
+ // Reset for next sample period
+ fpsHistory.clear();
+ fpsAccumulationTime = 0.0f;
+ }
+
ImGui::Text("Traced Depth %d", imguiData->TracedDepth);
ImGui::Text("Application average %.3f ms/frame (%.1f FPS)", 1000.0f / ImGui::GetIO().Framerate, ImGui::GetIO().Framerate);
ImGui::End();
@@ -349,9 +378,22 @@ int main(int argc, char** argv)
}
const char* sceneFile = argv[1];
+ const char* enviFile = argv[2];
+ const char* objFile = argv[3];
// Load scene file
scene = new Scene(sceneFile);
+ scene->loadEnvironmentMap(enviFile);
+
+ // load obj file
+ //scene->loadOBJ(objFile);
+
+ // BUILD BVH BASED ON GEOMS
+ scene->centroids = std::vector(scene->geoms.size(), glm::vec3());
+ for (int i = 0; i < scene->geoms.size(); i++) {
+ scene->centroids[i] = scene->geoms[i].translation;
+ }
+ scene->loadBVH();
//Create Instance for ImGUIData
guiData = new GuiDataContainer();
@@ -551,3 +593,16 @@ void mousePositionCallback(GLFWwindow* window, double xpos, double ypos)
lastX = xpos;
lastY = ypos;
}
+
+void scrollCallback(GLFWwindow* window, double xoffset, double yoffset)
+{
+ if (MouseOverImGuiWindow())
+ {
+ return; // Don't zoom if mouse is over ImGui
+ }
+
+ // Zoom in/out based on scroll direction
+ zoom -= (float)yoffset * 0.5f; // Adjust 0.1f to change zoom sensitivity
+ zoom = std::fmax(0.1f, zoom); // Prevent zooming too close
+ camchanged = true;
+}
\ No newline at end of file
diff --git a/src/pathtrace.cu b/src/pathtrace.cu
index 709c231ba..bd975c5ff 100644
--- a/src/pathtrace.cu
+++ b/src/pathtrace.cu
@@ -7,6 +7,15 @@
#include
#include
+#include
+#include
+#include
+#include
+
+#include
+#include
+#include
+
#include "sceneStructs.h"
#include "scene.h"
#include "glm/glm.hpp"
@@ -76,12 +85,22 @@ __global__ void sendImageToPBO(uchar4* pbo, glm::ivec2 resolution, int iter, glm
static Scene* hst_scene = NULL;
static GuiDataContainer* guiData = NULL;
static glm::vec3* dev_image = NULL;
+static glm::vec3* dev_bloomImage = NULL;
+static glm::vec3* dev_bloomMask = NULL;
+static glm::vec3* dev_bloomMaskBlur = NULL;
static Geom* dev_geoms = NULL;
static Material* dev_materials = NULL;
static PathSegment* dev_paths = NULL;
static ShadeableIntersection* dev_intersections = NULL;
-// TODO: static variables for device memory, any extra info you need, etc
-// ...
+static float* dev_EnviMap = NULL;
+static int* dev_matIds = NULL;
+
+static bvhNode* dev_bvhTree = NULL;
+static int* dev_bvhGeoIdx = NULL;
+
+thrust::device_ptr dev_thrust_matId;
+thrust::device_ptr dev_thrust_pathIdx;
+thrust::device_ptr dev_thrust_intersections;
void InitDataContainer(GuiDataContainer* imGuiData)
{
@@ -98,11 +117,26 @@ void pathtraceInit(Scene* scene)
cudaMalloc(&dev_image, pixelcount * sizeof(glm::vec3));
cudaMemset(dev_image, 0, pixelcount * sizeof(glm::vec3));
+ cudaMalloc(&dev_bloomImage, pixelcount * sizeof(glm::vec3));
+ cudaMemset(dev_bloomImage, 0, pixelcount * sizeof(glm::vec3));
+
+ cudaMalloc(&dev_bloomMask, pixelcount * sizeof(glm::vec3));
+ cudaMemset(dev_bloomMask, 0, pixelcount * sizeof(glm::vec3));
+
+ cudaMalloc(&dev_bloomMaskBlur, pixelcount * sizeof(glm::vec3));
+ cudaMemset(dev_bloomMaskBlur, 0, pixelcount * sizeof(glm::vec3));
+
cudaMalloc(&dev_paths, pixelcount * sizeof(PathSegment));
cudaMalloc(&dev_geoms, scene->geoms.size() * sizeof(Geom));
cudaMemcpy(dev_geoms, scene->geoms.data(), scene->geoms.size() * sizeof(Geom), cudaMemcpyHostToDevice);
+ cudaMalloc(&dev_bvhTree, scene->bvhTree.size() * sizeof(bvhNode));
+ cudaMemcpy(dev_bvhTree, scene->bvhTree.data(), scene->bvhTree.size() * sizeof(bvhNode), cudaMemcpyHostToDevice);
+
+ cudaMalloc(&dev_bvhGeoIdx, scene->bvhGeoIdx.size() * sizeof(int));
+ cudaMemcpy(dev_bvhGeoIdx, scene->bvhGeoIdx.data(), scene->bvhGeoIdx.size() * sizeof(int), cudaMemcpyHostToDevice);
+
cudaMalloc(&dev_materials, scene->materials.size() * sizeof(Material));
cudaMemcpy(dev_materials, scene->materials.data(), scene->materials.size() * sizeof(Material), cudaMemcpyHostToDevice);
@@ -110,6 +144,21 @@ void pathtraceInit(Scene* scene)
cudaMemset(dev_intersections, 0, pixelcount * sizeof(ShadeableIntersection));
// TODO: initialize any extra device memeory you need
+ cudaMalloc(&dev_matIds, pixelcount * sizeof(int));
+
+ if (hst_scene->enviMap != nullptr && hst_scene->enviMap->image != nullptr) {
+ int enviSize = hst_scene->enviMap->width * hst_scene->enviMap->height * hst_scene->enviMap->channels;
+
+ if (enviSize > 0) {
+ cudaMalloc(&dev_EnviMap, enviSize * sizeof(float));
+ cudaMemcpy(dev_EnviMap, hst_scene->enviMap->image, enviSize * sizeof(float), cudaMemcpyHostToDevice);
+ }
+
+ }
+ else {
+ dev_EnviMap = nullptr;
+ }
+
checkCUDAError("pathtraceInit");
}
@@ -117,15 +166,26 @@ void pathtraceInit(Scene* scene)
void pathtraceFree()
{
cudaFree(dev_image); // no-op if dev_image is null
+ cudaFree(dev_bloomImage);
+ cudaFree(dev_bloomMask);
cudaFree(dev_paths);
cudaFree(dev_geoms);
+ cudaFree(dev_bvhTree);
+ cudaFree(dev_bvhGeoIdx);
cudaFree(dev_materials);
cudaFree(dev_intersections);
// TODO: clean up any extra device memory you created
+ cudaFree(dev_matIds);
+
+ if (dev_EnviMap != nullptr) {
+ cudaFree(dev_EnviMap);
+ dev_EnviMap = nullptr;
+ }
checkCUDAError("pathtraceFree");
}
+
/**
* Generate PathSegments with rays from the camera through the screen into the
* scene, which is the first bounce of rays.
@@ -134,11 +194,81 @@ void pathtraceFree()
* motion blur - jitter rays "in time"
* lens effect - jitter ray origin positions based on a lens
*/
-__global__ void generateRayFromCamera(Camera cam, int iter, int traceDepth, PathSegment* pathSegments)
+
+__device__ glm::vec2 uniformDiskConcentric(const glm::vec2& sample)
+{
+
+ float M_PI = 3.14159265359;
+
+ float x = 2.0f * sample[0] - 1.0f;
+ float y = 2.0f * sample[1] - 1.0f;
+
+ float r = 1.f;
+ float a = 0.f;
+
+ if (x == 0 && y == 0) {
+ return glm::vec2(0.f, 0.f);
+ }
+ if (abs(x) > abs(y)) {
+ r *= x;
+ a = (M_PI * 0.25) * (y / x);
+ }
+ else {
+ r *= y;
+ a = (M_PI * 0.5) - ((M_PI * 0.25) * (x / y));
+ }
+
+ return glm::vec2(cos(a) * r, sin(a) * r);
+
+}
+
+__device__ glm::vec2 sampleSphericalMap(glm::vec3 v) {
+ v = glm::normalize(v);
+
+ // Convert to spherical coordinates
+ float theta = atan2f(v.x, v.z);
+ float phi = acosf(glm::clamp(v.y, -1.0f, 1.0f));
+
+ float u = (theta + PI) / (2.0f * PI);
+ float v_coord = phi / PI;
+
+ return glm::vec2(u, v_coord);
+}
+
+__device__ glm::vec3 sampleEnvironmentMap(const glm::vec3& direction, const float* envMap, const int width, const int height) {
+
+ // Convert direction to UV
+ glm::vec2 uv = sampleSphericalMap(direction);
+
+ // Convert UV to pixel coordinates
+ int x = (int)(uv.x * width) % width;
+ int y = (int)(uv.y * height) % height;
+
+ // Handle wrapping for x coordinate
+ if (x < 0) x += width;
+ if (y < 0) y = 0;
+ if (y >= height) y = height - 1;
+
+ // Calculate index (assuming RGB interleaved format)
+ int index = (y * width + x) * 3;
+
+ // Sample RGB values
+ return glm::vec3(
+ envMap[index], // R
+ envMap[index + 1], // G
+ envMap[index + 2] // B
+ );
+}
+
+__global__ void generateRayFromCamera(Camera cam, int iter, int traceDepth, PathSegment* pathSegments, float focalDist, float appature)
{
int x = (blockIdx.x * blockDim.x) + threadIdx.x;
int y = (blockIdx.y * blockDim.y) + threadIdx.y;
+ thrust::default_random_engine rng = makeSeededRandomEngine(iter, x * y, traceDepth);
+
+ thrust::uniform_real_distribution u01(-0.5, 0.5);
+
if (x < cam.resolution.x && y < cam.resolution.y) {
int index = x + (y * cam.resolution.x);
PathSegment& segment = pathSegments[index];
@@ -151,66 +281,195 @@ __global__ void generateRayFromCamera(Camera cam, int iter, int traceDepth, Path
- cam.right * cam.pixelLength.x * ((float)x - (float)cam.resolution.x * 0.5f)
- cam.up * cam.pixelLength.y * ((float)y - (float)cam.resolution.y * 0.5f)
);
+
+ // JITTER RAY FOR AA
+ segment.ray.direction += cam.right * cam.pixelLength.x * u01(rng);
+ segment.ray.direction += cam.up * cam.pixelLength.y * u01(rng);
+
+ // DEPTH OF FIELD / THIN LENSE CAMERA
+ float focalDistance = focalDist;
+ float lenseRad = appature;
+ thrust::uniform_real_distribution u01(0.0, 1.0);
+
+ glm::vec3 camForward = glm::normalize(cam.view - cam.position);
+
+ float denom = glm::dot(segment.ray.direction, camForward);
+ if (abs(denom) < 0.0001f) {
+ return;
+ }
+
+ float t = focalDistance / denom;
+ glm::vec3 pFocus_world = segment.ray.origin + t * segment.ray.direction;
+
+ glm::vec2 lensUV = uniformDiskConcentric(glm::vec2(u01(rng), u01(rng)));
+ lensUV *= lenseRad;
+
+ segment.ray.origin = cam.position + cam.right * lensUV.x + cam.up * lensUV.y;
+ segment.ray.direction = glm::normalize(pFocus_world - segment.ray.origin);
+
+
+ // JITTER RAY FOR AA
+ segment.ray.direction += cam.right * cam.pixelLength.x * u01(rng);
+ segment.ray.direction += cam.up * cam.pixelLength.y * u01(rng);
segment.pixelIndex = index;
segment.remainingBounces = traceDepth;
}
}
+__global__ void fillMaterialId(int num_paths, int* dev_matIds, ShadeableIntersection* dev_intersections) {
+ int path_index = blockIdx.x * blockDim.x + threadIdx.x;
+ if (path_index < num_paths)
+ {
+ dev_matIds[path_index] = dev_intersections[path_index].materialId;
+ }
+}
+
// TODO:
// computeIntersections handles generating ray intersections ONLY.
// Generating new rays is handled in your shader(s).
// Feel free to modify the code below.
+
+
+__device__ bool IntersectAABB(const Ray& ray, float t, const glm::vec3 bmin, const glm::vec3 bmax) {
+ float tx1 = (bmin.x - ray.origin.x) / ray.direction.x, tx2 = (bmax.x - ray.origin.x) / ray.direction.x;
+ float tmin = glm::min(tx1, tx2), tmax = glm::max(tx1, tx2);
+
+ float ty1 = (bmin.y - ray.origin.y) / ray.direction.y, ty2 = (bmax.y - ray.origin.y) / ray.direction.y;
+ tmin = glm::max(tmin, glm::min(ty1, ty2)), tmax = glm::min(tmax, glm::max(ty1, ty2));
+ float tz1 = (bmin.z - ray.origin.z) / ray.direction.z, tz2 = (bmax.z - ray.origin.z) / ray.direction.z;
+ tmin = glm::max(tmin, glm::min(tz1, tz2)), tmax = glm::min(tmax, glm::max(tz1, tz2));
+
+ return tmax >= tmin && tmin < t && tmax > 0;
+}
+
+__device__ float computeBVHintersectDist(int numGeos, int numNodes, Ray ray, float& t, Geom* geoms, bvhNode* bvhTree, int* bvhGeoIdx, const int nodeIndex,
+ glm::vec3& intersect_point, glm::vec3& normal, int& hit_geom_index) {
+
+
+ int stack[32];
+ int stackPtr = 0;
+ stack[stackPtr++] = numNodes - 1;
+
+ while (stackPtr > 0) {
+ int nodeIndex = stack[--stackPtr];
+ bvhNode curNode = bvhTree[nodeIndex];
+
+ if (!IntersectAABB(ray, t, curNode.aabbMin, curNode.aabbMax)) {
+ continue;
+ }
+
+ if (curNode.isLeaf == 1) {
+ glm::vec3 tmp_intersect;
+ glm::vec3 tmp_normal;
+ bool outside = true;
+ float tTest;
+
+ for (int i = curNode.startIdx; i < curNode.endIdx; i++) {
+ if (i < 0 || i >= numGeos) continue;
+ int geoIdx = bvhGeoIdx[i];
+ if (geoIdx < 0 || geoIdx >= numGeos) continue;
+
+ Geom& geom = geoms[geoIdx];
+
+ if (geom.type == SPHERE) {
+ tTest = sphereIntersectionTest(geom, ray, tmp_intersect, tmp_normal, outside);
+ }
+ else if (geom.type == CUBE) {
+ tTest = boxIntersectionTest(geom, ray, tmp_intersect, tmp_normal, outside);
+ }
+ else if (geom.type == DISK) {
+ tTest = diskIntersectionTest(geom, ray, tmp_intersect, tmp_normal);
+ }
+ else if (geom.type == TRIANGLE) {
+ tTest = triangleIntersectionTest(geom, ray, tmp_intersect, tmp_normal, outside);
+ }
+
+ if (tTest > 0.0f && tTest < t) {
+ t = tTest;
+ hit_geom_index = geoIdx;
+ intersect_point = tmp_intersect;
+ normal = tmp_normal;
+ }
+ }
+ }
+ else {
+ if (curNode.rightChild >= 0) {
+ stack[stackPtr++] = curNode.rightChild;
+ }
+ if (curNode.leftChild >= 0) {
+ stack[stackPtr++] = curNode.leftChild;
+ }
+ }
+ }
+
+ return t;
+
+}
+
__global__ void computeIntersections(
int depth,
int num_paths,
PathSegment* pathSegments,
Geom* geoms,
+ bvhNode* bvhTree,
+ int* bvhGeoIdx,
int geoms_size,
+ int bvhTree_size,
ShadeableIntersection* intersections)
-{
+{
+
int path_index = blockIdx.x * blockDim.x + threadIdx.x;
if (path_index < num_paths)
{
PathSegment pathSegment = pathSegments[path_index];
- float t;
+ // FOR GOING BACK TO NAIVE INTERSECTIONS SWITCH t AND t_min
+ float t = FLT_MAX;
glm::vec3 intersect_point;
glm::vec3 normal;
- float t_min = FLT_MAX;
+ float t_min;
int hit_geom_index = -1;
bool outside = true;
glm::vec3 tmp_intersect;
glm::vec3 tmp_normal;
- // naive parse through global geoms
-
- for (int i = 0; i < geoms_size; i++)
- {
- Geom& geom = geoms[i];
- if (geom.type == CUBE)
- {
- t = boxIntersectionTest(geom, pathSegment.ray, tmp_intersect, tmp_normal, outside);
- }
- else if (geom.type == SPHERE)
- {
- t = sphereIntersectionTest(geom, pathSegment.ray, tmp_intersect, tmp_normal, outside);
- }
- // TODO: add more intersection tests here... triangle? metaball? CSG?
-
- // Compute the minimum t from the intersection tests to determine what
- // scene geometry object was hit first.
- if (t > 0.0f && t_min > t)
- {
- t_min = t;
- hit_geom_index = i;
- intersect_point = tmp_intersect;
- normal = tmp_normal;
- }
- }
+ //
+ //for (int i = 0; i < geoms_size; i++)
+ //{
+ // Geom& geom = geoms[i];
+ //
+ // if (geom.type == CUBE)
+ // {
+ // t = boxIntersectionTest(geom, pathSegment.ray, tmp_intersect, tmp_normal, outside);
+ // }
+ // else if (geom.type == SPHERE)
+ // {
+ // t = sphereIntersectionTest(geom, pathSegment.ray, tmp_intersect, tmp_normal, outside);
+ // }
+ // else if (geom.type == DISK) {
+ // t = diskIntersectionTest(geom, pathSegment.ray, tmp_intersect, tmp_normal);
+ // }
+ // else if (geom.type == TRIANGLE) {
+ // t = triangleIntersectionTest(geom, pathSegment.ray, tmp_intersect, tmp_normal, outside);
+ // }
+ // // TODO: add more intersection tests here... triangle? metaball? CSG?
+ //
+ // // Compute the minimum t from the intersection tests to determine what
+ // // scene geometry object was hit first.
+ // if (t > 0.0f && t_min > t)
+ // {
+ // t_min = t;
+ // hit_geom_index = i;
+ // intersect_point = tmp_intersect;
+ // normal = tmp_normal;
+ // }
+ //}
+
+ t_min = computeBVHintersectDist(geoms_size, bvhTree_size, pathSegment.ray, t, geoms, bvhTree, bvhGeoIdx, 0, intersect_point, normal, hit_geom_index);
if (hit_geom_index == -1)
{
@@ -221,11 +480,13 @@ __global__ void computeIntersections(
// The ray hits something
intersections[path_index].t = t_min;
intersections[path_index].materialId = geoms[hit_geom_index].materialid;
+ intersections[path_index].geoId = hit_geom_index;
intersections[path_index].surfaceNormal = normal;
}
}
}
+
// LOOK: "fake" shader demonstrating what you might do with the info in
// a ShadeableIntersection, as well as how to use thrust's random number
// generator. Observe that since the thrust random number generator basically
@@ -280,6 +541,129 @@ __global__ void shadeFakeMaterial(
}
}
+__global__ void mirrorShader(int iter,
+ int num_paths,
+ ShadeableIntersection* shadeableIntersections,
+ PathSegment* pathSegments,
+ Material* materials) {
+
+ int idx = blockIdx.x * blockDim.x + threadIdx.x;
+ if (idx < num_paths)
+ {
+ ShadeableIntersection intersection = shadeableIntersections[idx];
+ if (intersection.t > 0.0f)
+ {
+ Material material = materials[intersection.materialId];
+ glm::vec3 materialColor = material.color;
+
+ thrust::default_random_engine rng = makeSeededRandomEngine(iter, idx, 8 - pathSegments[idx].remainingBounces);
+ thrust::uniform_real_distribution u01(0, 1);
+
+ if (material.emittance > 0.0f) {
+ pathSegments[idx].color *= (materialColor * material.emittance);
+ return;
+ }
+ else {
+
+ pathSegments[idx].color *= material.color;
+
+ if (pathSegments[idx].remainingBounces <= 0) {
+ pathSegments[idx].color = glm::vec3(0.0f);
+ return;
+ }
+ else {
+ glm::vec3 magic = getPointOnRay(pathSegments[idx].ray, intersection.t);
+ scatterRay(pathSegments[idx], magic, intersection.surfaceNormal, material, rng);
+ }
+ }
+ }
+ else {
+ pathSegments[idx].color = glm::vec3(0.0f);
+ }
+ }
+
+}
+
+__global__ void shadeRay(int iter,
+ int num_paths,
+ ShadeableIntersection* shadeableIntersections,
+ PathSegment* pathSegments,
+ Material* materials,
+ int depth,
+ Geom* geoms,
+ int geoms_size,
+ float* enviMap, int enviWidth, int enviHeight) {
+
+ float pi = 3.14159265359;
+ float INV_PI = 1.f / pi;
+ int idx = blockIdx.x * blockDim.x + threadIdx.x;
+
+ if (idx < num_paths)
+ {
+ if (pathSegments[idx].remainingBounces < 0) {
+ return;
+ }
+ ShadeableIntersection intersection = shadeableIntersections[idx];
+ if (intersection.t > 0.0f)
+ {
+
+ thrust::default_random_engine rng = makeSeededRandomEngine(iter, idx, depth);
+ //thrust::uniform_real_distribution u01(0, 1);
+
+ Material material = materials[intersection.materialId];
+ glm::vec3 materialColor = material.color;
+
+ if (intersection.materialId == 0) {
+ Geom geo = geoms[intersection.geoId];
+ glm::vec3 magic = getPointOnRay(pathSegments[idx].ray, intersection.t);
+ blackHoleRay(pathSegments[idx], magic, intersection.surfaceNormal, geo.invTranspose, material, rng);
+
+ }
+ else if (material.emittance > 0.0f) {
+ pathSegments[idx].color *= materialColor * material.emittance;
+ pathSegments[idx].remainingBounces = -1;
+ }
+ else {
+ pathSegments[idx].color *= materialColor;
+ glm::vec3 magic = getPointOnRay(pathSegments[idx].ray, intersection.t);
+ scatterRay(pathSegments[idx], magic, intersection.surfaceNormal, material, rng);
+
+ /*Ray lightDir = Ray();
+ lightDir.direction = -glm::normalize(glm::vec3(0.0, 10.0, 0.0) - magic);
+ lightDir.origin = magic + intersection.surfaceNormal * 0.01f;
+ float lightDist = glm::length(glm::vec3(0.0, 10.0, 0.0) - magic);
+ float intersectDist = computeIntersectDist(lightDir, geoms, geoms_size);
+
+ if (abs(lightDist - intersectDist) < 0.01) {
+ pathSegments[idx].color *= glm::vec3(1.0) * 5.f;
+ }*/
+
+ }
+ }
+ else {
+ glm::vec3 enviColor = sampleEnvironmentMap(pathSegments[idx].ray.direction, enviMap, enviWidth, enviHeight);
+ pathSegments[idx].color *= enviColor;
+ pathSegments[idx].remainingBounces = -1;
+ }
+ }
+}
+
+// STREAM COMPACTION BOOL
+struct IsAlive {
+ __host__ __device__
+ bool operator()(const PathSegment& pathSeg) {
+ return pathSeg.remainingBounces >= 0;
+ }
+};
+
+struct IsAliveZip {
+ __host__ __device__
+ bool operator()(const thrust::tuple& t) {
+ IsAlive isAlive;
+ return isAlive(thrust::get<0>(t));
+ }
+};
+
// Add the current iteration's output to the overall image
__global__ void finalGather(int nPaths, glm::vec3* image, PathSegment* iterationPaths)
{
@@ -292,6 +676,126 @@ __global__ void finalGather(int nPaths, glm::vec3* image, PathSegment* iteration
}
}
+__global__ void bloomHighPass(int nPaths, PathSegment* iterationPaths, glm::vec3* image, glm::vec3* bloomMask, glm::ivec2 resolution, int iter, float thresh) {
+ int x = (blockIdx.x * blockDim.x) + threadIdx.x;
+ int y = (blockIdx.y * blockDim.y) + threadIdx.y;
+
+ if (x < resolution.x && y < resolution.y)
+ {
+ int index = x + (y * resolution.x);
+ glm::vec3 color = image[index];
+ //scolor /= float(iter);
+
+ float brightness = glm::dot(color, glm::vec3(0.2126f, 0.7152f, 0.0722f));
+ if (brightness > thresh) {
+ bloomMask[index] = color;
+ }
+ else {
+ bloomMask[index] = glm::vec3(0.0f);
+ }
+ }
+}
+
+__global__ void bloomBlurY(int nPaths, glm::vec3* ibloomMask, glm::vec3* oBloomMask, int imageWidth, int imageHeight) {
+ int x = (blockIdx.x * blockDim.x) + threadIdx.x;
+ int y = (blockIdx.y * blockDim.y) + threadIdx.y;
+
+ float weight[11] = {
+ 0.1247f, // center
+ 0.1226f, // ±1
+ 0.1167f, // ±2
+ 0.1075f, // ±3
+ 0.0957f, // ±4
+ 0.0818f, // ±5
+ 0.0672f, // ±6
+ 0.0530f, // ±7
+ 0.0399f, // ±8
+ 0.0285f, // ±9
+ 0.0193f // ±10
+ };
+
+ if (x < imageWidth && y < imageHeight)
+ {
+ int index = x + (y * imageWidth);
+ glm::vec3 result = ibloomMask[index] * weight[0];
+ for (int i = 1; i < 11; ++i)
+ {
+ int posIdx = x + ((y + i) * imageWidth);
+ int negIdx = x + ((y - i) * imageWidth);
+
+ if ((y + i) < imageHeight) {
+ result += ibloomMask[posIdx] * weight[i];
+ }
+ if ((y - i) >= 0) {
+ result += ibloomMask[negIdx] * weight[i];
+ }
+
+ }
+
+ oBloomMask[index] = result;
+ }
+}
+
+__global__ void bloomBlurX(int nPaths, glm::vec3* ibloomMask, glm::vec3* oBloomMask, int imageWidth, int imageHeight) {
+ int x = (blockIdx.x * blockDim.x) + threadIdx.x;
+ int y = (blockIdx.y * blockDim.y) + threadIdx.y;
+
+ float weight[11] = {
+ 0.1247f, // centers
+ 0.1226f, // ±1
+ 0.1167f, // ±2
+ 0.1075f, // ±3
+ 0.0957f, // ±4
+ 0.0818f, // ±5
+ 0.0672f, // ±6
+ 0.0530f, // ±7
+ 0.0399f, // ±8
+ 0.0285f, // ±9
+ 0.0193f // ±10
+ };
+
+ if (x < imageWidth && y < imageHeight)
+ {
+ int index = x + (y * imageWidth);
+ glm::vec3 result = ibloomMask[index] * weight[0];
+ for (int i = 1; i < 11; ++i)
+ {
+ int posIdx = (x + i) + (y * imageWidth);
+ int negIdx = (x - i) + (y * imageWidth);
+
+ if ((x + i) < imageWidth) {
+ result += ibloomMask[posIdx] * weight[i];
+ }
+ if ((x - i) >= 0) {
+ result += ibloomMask[negIdx] * weight[i];
+ }
+ }
+
+ oBloomMask[index] = result;
+ }
+}
+
+__global__ void bloomBlend(int nPaths, glm::vec3* bloomMask, glm::vec3* image, glm::vec3* dev_bloomImage, int imageWidth, int imageHeight) {
+ int x = (blockIdx.x * blockDim.x) + threadIdx.x;
+ int y = (blockIdx.y * blockDim.y) + threadIdx.y;
+ int index = x + (y * imageWidth);
+
+ if (x < imageWidth && y < imageHeight)
+ {
+ const float gamma = 2.2;
+ const float exposure = 0.7;
+ glm::vec3 hdrColor = image[index];
+ glm::vec3 bloomColor = bloomMask[index];
+ hdrColor += bloomColor;
+
+ glm::vec3 result = glm::vec3(1.0f) - exp(-hdrColor * exposure);
+ // also gamma correct while we're at it
+ result = pow(result, glm::vec3(1.0f / gamma));
+
+ dev_bloomImage[index] = glm::vec3(hdrColor);
+ }
+}
+
/**
* Wrapper for the __global__ call that sets up the kernel calls and does a ton
* of memory management
@@ -342,16 +846,25 @@ void pathtrace(uchar4* pbo, int frame, int iter)
// TODO: perform one iteration of path tracing
- generateRayFromCamera<<>>(cam, iter, traceDepth, dev_paths);
+ generateRayFromCamera<<>>(cam, iter, traceDepth, dev_paths, 36.0, 0.0);
checkCUDAError("generate camera ray");
int depth = 0;
PathSegment* dev_path_end = dev_paths + pixelcount;
int num_paths = dev_path_end - dev_paths;
+ int og_num_paths = num_paths;
+
+ // ENVI MAP VARIABLES:
+ int hdrHeight = hst_scene->enviMap->height;
+ int hdrWidth = hst_scene->enviMap->width;
+
+ if (dev_bvhTree == NULL) {
+ printf("ERROR: dev_bvhTree is NULL!\n");
+ }
+
// --- PathSegment Tracing Stage ---
// Shoot ray into scene, bounce between objects, push shading chunks
-
bool iterationComplete = false;
while (!iterationComplete)
{
@@ -365,12 +878,14 @@ void pathtrace(uchar4* pbo, int frame, int iter)
num_paths,
dev_paths,
dev_geoms,
+ dev_bvhTree,
+ dev_bvhGeoIdx,
hst_scene->geoms.size(),
+ hst_scene->bvhTree.size(),
dev_intersections
);
checkCUDAError("trace one bounce");
cudaDeviceSynchronize();
- depth++;
// TODO:
// --- Shading Stage ---
@@ -381,15 +896,60 @@ void pathtrace(uchar4* pbo, int frame, int iter)
// TODO: compare between directly shading the path segments and shading
// path segments that have been reshuffled to be contiguous in memory.
- shadeFakeMaterial<<>>(
+ //shadeFakeMaterial<<>>(
+ // iter,
+ // num_paths,
+ // dev_intersections,
+ // dev_paths,
+ // dev_materials
+ //);
+ //iterationComplete = true; // TODO: should be based off stream compaction results.
+
+
+ // SORT RAYS BY MATERIAL ID
+ //fillMaterialId << > > (
+ // num_paths, dev_matIds, dev_intersections
+ //);
+
+ //dev_thrust_matId = thrust::device_pointer_cast(dev_matIds);
+ //dev_thrust_pathIdx = thrust::device_pointer_cast(dev_paths);
+ //dev_thrust_intersections = thrust::device_pointer_cast(dev_intersections);
+
+ //thrust::sort_by_key(
+ // dev_thrust_matId,
+ // dev_thrust_matId + num_paths,
+ // thrust::make_zip_iterator(thrust::make_tuple(
+ // dev_thrust_pathIdx,
+ // dev_thrust_intersections
+ // ))
+ //);
+
+ shadeRay << < numblocksPathSegmentTracing, blockSize1d >> > (
iter,
num_paths,
dev_intersections,
dev_paths,
- dev_materials
+ dev_materials,
+ depth,
+ dev_geoms,
+ hst_scene->geoms.size(),
+ dev_EnviMap, hdrWidth, hdrHeight
);
- iterationComplete = true; // TODO: should be based off stream compaction results.
+ // STREAM COMPACTION
+ auto zip_begin = thrust::make_zip_iterator(thrust::make_tuple(dev_paths, dev_intersections));
+ auto zip_end = thrust::make_zip_iterator(thrust::make_tuple(dev_paths + num_paths, dev_intersections + num_paths));
+ auto partition_point = thrust::partition(thrust::device,
+ zip_begin, zip_end,
+ IsAliveZip{});
+ num_paths = partition_point - zip_begin;
+
+ if (depth > traceDepth || num_paths <= 0) {
+ iterationComplete = true; // TODO: should be based off stream compaction results.
+ }
+
+ depth++;
+
if (guiData != NULL)
{
guiData->TracedDepth = depth;
@@ -398,16 +958,24 @@ void pathtrace(uchar4* pbo, int frame, int iter)
// Assemble this iteration and apply it to the image
dim3 numBlocksPixels = (pixelcount + blockSize1d - 1) / blockSize1d;
- finalGather<<>>(num_paths, dev_image, dev_paths);
+ finalGather<<>>(pixelcount, dev_image, dev_paths);
///////////////////////////////////////////////////////////////////////////
+
+ // BLOOM POST PROCESS
+
+ cudaMemset(dev_bloomMask, 0, pixelcount * sizeof(glm::vec3));
+ bloomHighPass << > > (pixelcount, dev_paths, dev_image, dev_bloomMask, cam.resolution, iter, 1.0f);
+ bloomBlurY << > > (pixelcount, dev_bloomMask, dev_bloomMaskBlur, cam.resolution.x, cam.resolution.y);
+ bloomBlurX << > > (pixelcount, dev_bloomMaskBlur, dev_bloomMask, cam.resolution.x, cam.resolution.y);
+ bloomBlend << > > (pixelcount, dev_bloomMask, dev_image, dev_bloomImage, cam.resolution.x, cam.resolution.y);
+ sendImageToPBO<<>>(pbo, cam.resolution, iter, dev_bloomImage);
// Send results to OpenGL buffer for rendering
- sendImageToPBO<<>>(pbo, cam.resolution, iter, dev_image);
+ //sendImageToPBO<<>>(pbo, cam.resolution, iter, dev_image);
// Retrieve image from GPU
- cudaMemcpy(hst_scene->state.image.data(), dev_image,
- pixelcount * sizeof(glm::vec3), cudaMemcpyDeviceToHost);
+ cudaMemcpy(hst_scene->state.image.data(), dev_bloomImage, pixelcount * sizeof(glm::vec3), cudaMemcpyDeviceToHost);
checkCUDAError("pathtrace");
}
diff --git a/src/scene.cpp b/src/scene.cpp
index 07ef304f8..593822268 100644
--- a/src/scene.cpp
+++ b/src/scene.cpp
@@ -5,6 +5,7 @@
#include
#include
#include "json.hpp"
+#include "tiny_obj_loader.h"
#include
#include
@@ -31,6 +32,289 @@ Scene::Scene(string filename)
}
}
+void Scene::loadEnvironmentMap(const std::string& hdrName)
+{
+ int width, height, channels;
+ float* data = stbi_loadf(hdrName.c_str(), &width, &height, &channels, 3);
+
+ enviMap = new EnvironmentMap();
+ enviMap->width = width;
+ enviMap->height = height;
+ enviMap->channels = channels;
+ enviMap->image = data;
+}
+
+//bvhNode Scene::buildTree(glm::vec3 min, glm::vec3 max, int idxStart, int idxEnd , int leafSize) {
+// bvhNode curNode;
+// curNode.aabbMax = max;
+// curNode.aabbMin = min;
+// curNode.startIdx = idxStart;
+// curNode.endIdx = idxEnd;
+// curNode.isLeaf = false;
+//
+// // BUILD BOUNDING BOX
+// for (int i = idxStart; i < idxEnd; i++) {
+// Geom geo = geoms[bvhGeoIdx[i]];
+// curNode.aabbMax.x = glm::max(curNode.aabbMax.x, geo.translation.x);
+// curNode.aabbMax.y = glm::max(curNode.aabbMax.y, geo.translation.y);
+// curNode.aabbMax.z = glm::max(curNode.aabbMax.z, geo.translation.z);
+//
+// curNode.aabbMin.x = glm::min(curNode.aabbMin.x, geo.translation.x);
+// curNode.aabbMin.y = glm::min(curNode.aabbMin.y, geo.translation.y);
+// curNode.aabbMin.z = glm::min(curNode.aabbMin.z, geo.translation.z);
+// }
+//
+// if (curNode.endIdx - curNode.startIdx <= leafSize) { // STOP SUBDIV
+// curNode.isLeaf = true;
+// curNode.leftChild = nullptr;
+// curNode.rightChild = nullptr;
+// bvhTree.push_back(curNode);
+// return curNode;
+// }
+// else {
+// float centerPos = 0.0;
+// printf("Max = [%f, %f, %f]\n", curNode.aabbMax.x, curNode.aabbMax.y, curNode.aabbMax.z);
+// printf("Min = [%f, %f, %f]\n", curNode.aabbMin.x, curNode.aabbMin.y, curNode.aabbMin.z);
+// glm::vec3 ext = curNode.aabbMax - curNode.aabbMin;
+// int maxis = 0;
+// if (ext.y > ext.x) maxis = 1;
+// if (ext.z > ext[maxis]) maxis = 2;
+// centerPos = 0.5f * (curNode.aabbMax[maxis] + curNode.aabbMin[maxis]);
+//
+// // Partition current region and recurse
+// std::vector firstHalf = {};
+// std::vector lastHalf = {};
+// for (int i = idxStart; i < idxEnd; i++) {
+// if (geoms[bvhGeoIdx.at(i)].translation[maxis] < centerPos) {
+// firstHalf.push_back(bvhGeoIdx.at(i));
+// }
+// else {
+// lastHalf.push_back(bvhGeoIdx.at(i));
+// }
+// }
+//
+//
+// printf("First Half: ");
+// for (int i = 0; i < firstHalf.size(); i++) {
+// printf("%d, ", firstHalf.at(i));
+// }
+// printf("\n");
+// printf("Last Half: ");
+// for (int i = 0; i < lastHalf.size(); i++) {
+// printf("%d, ", lastHalf.at(i));
+// }
+// printf("\n");
+//
+// ///////////
+// int midIdx = idxStart + firstHalf.size();
+// firstHalf.insert(firstHalf.end(), lastHalf.begin(), lastHalf.end());
+// //////////////
+//
+// printf("New First Half: ");
+// for (int i = 0; i < firstHalf.size(); i++) {
+// printf("%d, ", firstHalf.at(i));
+// }
+// printf("\n");
+//
+// //////////
+// for (int i = idxStart; i <= idxEnd; i++) {
+// bvhGeoIdx.at(i) = firstHalf.at(i - idxStart);
+// }
+// /////////////////
+//
+// printf("Size = %d, front = %d, last = %d \n", bvhGeoIdx.size(), firstHalf.size(), lastHalf.size());
+// for (int i = 0; i < bvhGeoIdx.size(); i++) {
+// printf("%d, ", bvhGeoIdx.at(i));
+// }
+// printf("\n");
+// printf("mid id = %d \n", midIdx);
+//
+// glm::vec3 centerUpBound = curNode.aabbMax;
+// centerUpBound[maxis] = centerPos;
+// curNode.leftChild = &buildTree(curNode.aabbMin, centerUpBound, idxStart, midIdx-1, leafSize);
+//
+// glm::vec3 centerLowBound = curNode.aabbMin;
+// centerLowBound[maxis] = centerPos;
+// curNode.rightChild = &buildTree(centerLowBound, curNode.aabbMax, midIdx, idxEnd, leafSize);
+// }
+//
+// bvhTree.push_back(curNode);
+// return curNode;
+//}
+
+
+
+bvhNode Scene::buildTree(int idxStart, int idxEnd, int leafSize) {
+ bvhNode curNode;
+ curNode.aabbMin = glm::vec3(+FLT_MAX);
+ curNode.aabbMax = glm::vec3(-FLT_MAX);
+ curNode.startIdx = idxStart;
+ curNode.endIdx = idxEnd;
+ curNode.isLeaf = 0;
+ curNode.leftChild = -1;
+ curNode.rightChild = -1;
+
+ // CALCULATE BOUNDS
+ for (int i = idxStart; i < idxEnd; i++) {
+ Geom curGeo = geoms[bvhGeoIdx[i]];
+ glm::vec3 translate = curGeo.translation;
+ glm::vec3 scale = curGeo.scale;
+
+ curNode.aabbMax.x = glm::max(curNode.aabbMax.x, translate.x + scale.x);
+ curNode.aabbMax.y = glm::max(curNode.aabbMax.y, translate.y + scale.y);
+ curNode.aabbMax.z = glm::max(curNode.aabbMax.z, translate.z + scale.z);
+
+ curNode.aabbMin.x = glm::min(curNode.aabbMin.x, translate.x - scale.x);
+ curNode.aabbMin.y = glm::min(curNode.aabbMin.y, translate.y - scale.y);
+ curNode.aabbMin.z = glm::min(curNode.aabbMin.z, translate.z - scale.z);
+
+ this->centroids[bvhGeoIdx[i]] = curGeo.translation;
+ }
+
+ if (curNode.endIdx - curNode.startIdx <= leafSize) { // STOP SUBDIV
+ curNode.isLeaf = 1;
+ curNode.leftChild = -1;
+ curNode.rightChild = -1;
+ return curNode;
+ }
+ else {
+ curNode.isLeaf = 0;
+
+ // CALCULATE MID POINT
+ glm::vec3 ext = curNode.aabbMax - curNode.aabbMin;
+ int maxis = 0;
+ if (ext.y > ext.x) maxis = 1;
+ if (ext.z > ext[maxis]) maxis = 2;
+
+ float midPos = (curNode.aabbMax[maxis] + curNode.aabbMin[maxis]) * 0.5;
+
+ auto first = bvhGeoIdx.begin() + idxStart;
+ auto last = bvhGeoIdx.begin() + idxEnd;
+
+ auto midIt = std::partition(first, last, [&](int primId) {
+ float c = this->centroids[primId][maxis];
+ return c <= midPos;
+ });
+ int mid = int(midIt - bvhGeoIdx.begin());
+
+ if (mid == idxStart || mid == idxEnd) {
+ mid = (idxStart + idxEnd) / 2;
+ auto first = bvhGeoIdx.begin() + idxStart;
+ auto nth = bvhGeoIdx.begin() + mid;
+ auto last = bvhGeoIdx.begin() + idxEnd;
+
+ std::nth_element(first, nth, last, [&](int ia, int ib) {
+ const Geom& A = geoms[ia];
+ const Geom& B = geoms[ib];
+ float ca = A.translation[maxis];
+ float cb = B.translation[maxis];
+ return ca < cb;
+ });
+ }
+
+ //for (int i = 0; i < bvhGeoIdx.size(); i++) {
+ // printf("%d, ", bvhGeoIdx.at(i));
+ //}
+ //printf("\n");
+
+ bvhNode leftChild = buildTree(idxStart, mid, leafSize);
+ bvhNode rightChild = buildTree(mid, idxEnd, leafSize);
+
+ curNode.leftChild = bvhTree.size();
+ bvhTree.push_back(leftChild);
+
+ curNode.rightChild = bvhTree.size();
+ bvhTree.push_back(rightChild);
+ }
+
+ return curNode;
+}
+
+void Scene::loadBVH() {
+ this->bvhTree = {};
+ if (geoms.size() == 0) {
+ return;
+ }
+
+ root = buildTree(0, bvhGeoIdx.size(), 8);
+ bvhTree.push_back(root);
+ this->root = root;
+}
+
+void Scene::loadOBJ(const std::string& objName) {
+ tinyobj::ObjReader reader;
+ tinyobj::ObjReaderConfig reader_config;
+ reader_config.triangulate = true;
+
+ if (!reader.ParseFromFile(objName, reader_config)) {
+ if (!reader.Error().empty()) {
+ std::cerr << "TinyObjReader: " << reader.Error();
+ }
+ exit(1);
+ }
+
+ if (!reader.Warning().empty()) {
+ std::cout << "TinyObjReader: " << reader.Warning();
+ }
+
+ auto& attrib = reader.GetAttrib();
+ auto& shapes = reader.GetShapes();
+
+ for (size_t s = 0; s < shapes.size(); s++) {
+ // Loop over faces(polygon)
+ size_t index_offset = 0;
+ for (size_t f = 0; f < shapes[s].mesh.num_face_vertices.size(); f++) {
+ size_t fv = size_t(shapes[s].mesh.num_face_vertices[f]);
+ Geom newTri;
+
+ tinyobj::index_t idx = shapes[s].mesh.indices[index_offset];
+ newTri.triPos1 = glm::vec3(
+ attrib.vertices[3 * idx.vertex_index + 0],
+ attrib.vertices[3 * idx.vertex_index + 1],
+ attrib.vertices[3 * idx.vertex_index + 2]
+ );
+
+ idx = shapes[s].mesh.indices[index_offset + 1];
+ newTri.triPos2 = glm::vec3(
+ attrib.vertices[3 * idx.vertex_index + 0],
+ attrib.vertices[3 * idx.vertex_index + 1],
+ attrib.vertices[3 * idx.vertex_index + 2]
+ );
+
+ idx = shapes[s].mesh.indices[index_offset + 2];
+ newTri.triPos3 = glm::vec3(
+ attrib.vertices[3 * idx.vertex_index + 0],
+ attrib.vertices[3 * idx.vertex_index + 1],
+ attrib.vertices[3 * idx.vertex_index + 2]
+ );
+
+ glm::vec3 centroid = newTri.triPos1 + newTri.triPos2 + newTri.triPos3;
+ centroid /= 3.0f;
+
+ float pos1Rad = glm::length(newTri.triPos1 - centroid);
+ float pos2Rad = glm::length(newTri.triPos2 - centroid);
+ float pos3Rad = glm::length(newTri.triPos3 - centroid);
+
+ float rad = glm::max(pos3Rad, glm::max(pos1Rad, pos2Rad));
+
+ index_offset += fv;
+
+ newTri.type = TRIANGLE;
+ newTri.materialid = 1;
+ newTri.translation = centroid;
+ newTri.rotation = glm::vec3(0.0);
+ newTri.scale = glm::vec3(rad);
+ newTri.transform = utilityCore::buildTransformationMatrix(
+ newTri.translation, newTri.rotation, newTri.scale);
+ newTri.inverseTransform = glm::inverse(newTri.transform);
+ newTri.invTranspose = glm::inverseTranspose(newTri.transform);
+
+ this->geoms.push_back(newTri);
+ this->bvhGeoIdx.push_back(this->bvhGeoIdx.size()-1);
+ }
+ }
+}
+
void Scene::loadFromJSON(const std::string& jsonName)
{
std::ifstream f(jsonName);
@@ -47,22 +331,46 @@ void Scene::loadFromJSON(const std::string& jsonName)
{
const auto& col = p["RGB"];
newMaterial.color = glm::vec3(col[0], col[1], col[2]);
+ const auto& spec = p["ROUGHNESS"];
+ newMaterial.specular.exponent = spec;
+ newMaterial.specular.color = glm::vec3(col[0], col[1], col[2]);
}
else if (p["TYPE"] == "Emitting")
{
const auto& col = p["RGB"];
newMaterial.color = glm::vec3(col[0], col[1], col[2]);
newMaterial.emittance = p["EMITTANCE"];
+
+ const auto& spec = p["ROUGHNESS"];
+ newMaterial.specular.exponent = spec;
+ newMaterial.specular.color = glm::vec3(col[0], col[1], col[2]);
}
else if (p["TYPE"] == "Specular")
{
+ const auto& col = p["RGB"];
+ const auto& spec = p["ROUGHNESS"];
+ newMaterial.color = glm::vec3(col[0], col[1], col[2]);
+ newMaterial.specular.exponent = spec;
+ newMaterial.specular.color = glm::vec3(col[0], col[1], col[2]);
+ }
+ else if (p["TYPE"] == "BlackHole")
+ {
+ const auto& innerRad = p["INNERRAD"];
+ const auto& outerRad = p["OUTERRAD"];
const auto& col = p["RGB"];
newMaterial.color = glm::vec3(col[0], col[1], col[2]);
+ newMaterial.blackHole.iRad = innerRad;
+ newMaterial.blackHole.oRad = outerRad;
+ newMaterial.emittance = p["EMITTANCE"];
}
MatNameToID[name] = materials.size();
materials.emplace_back(newMaterial);
}
+
+ // READ GEOMETRY
const auto& objectsData = data["Objects"];
+ bvhGeoIdx = {};
+ int idx = 0;
for (const auto& p : objectsData)
{
const auto& type = p["TYPE"];
@@ -71,10 +379,14 @@ void Scene::loadFromJSON(const std::string& jsonName)
{
newGeom.type = CUBE;
}
- else
+ else if (type == "sphere")
{
newGeom.type = SPHERE;
}
+ else
+ {
+ newGeom.type = DISK;
+ }
newGeom.materialid = MatNameToID[p["MATERIAL"]];
const auto& trans = p["TRANS"];
const auto& rotat = p["ROTAT"];
@@ -88,7 +400,10 @@ void Scene::loadFromJSON(const std::string& jsonName)
newGeom.invTranspose = glm::inverseTranspose(newGeom.transform);
geoms.push_back(newGeom);
+ bvhGeoIdx.push_back(idx);
+ idx++;
}
+
const auto& cameraData = data["Camera"];
Camera& camera = state.camera;
RenderState& state = this->state;
@@ -122,3 +437,4 @@ void Scene::loadFromJSON(const std::string& jsonName)
state.image.resize(arraylen);
std::fill(state.image.begin(), state.image.end(), glm::vec3());
}
+
diff --git a/src/scene.h b/src/scene.h
index 8bd2b96f0..0ca878635 100644
--- a/src/scene.h
+++ b/src/scene.h
@@ -3,14 +3,29 @@
#include "sceneStructs.h"
#include
+//#define STB_IMAGE_IMPLEMENTATION
+#include "stb_image.h"
+
class Scene
{
private:
void loadFromJSON(const std::string& jsonName);
public:
Scene(std::string filename);
+ void loadEnvironmentMap(const std::string& hdrName);
+ void loadBVH();
+ bvhNode buildTree(int idxStart, int idxEndl, int leafSize);
+ void loadOBJ(const std::string& objName);
std::vector geoms;
+ std::vector centroids;
std::vector materials;
RenderState state;
+ EnvironmentMap* enviMap;
+
+ // BVH STRUCTURE
+ std::vector bvhTree;
+ std::vector bvhNodeIndex;
+ std::vector bvhGeoIdx;
+ bvhNode root;
};
diff --git a/src/sceneStructs.h b/src/sceneStructs.h
index b7a4f7352..e56e33000 100644
--- a/src/sceneStructs.h
+++ b/src/sceneStructs.h
@@ -12,7 +12,9 @@
enum GeomType
{
SPHERE,
- CUBE
+ CUBE,
+ DISK,
+ TRIANGLE
};
struct Ray
@@ -25,6 +27,11 @@ struct Geom
{
enum GeomType type;
int materialid;
+
+ glm::vec3 triPos1;
+ glm::vec3 triPos2;
+ glm::vec3 triPos3;
+
glm::vec3 translation;
glm::vec3 rotation;
glm::vec3 scale;
@@ -33,6 +40,18 @@ struct Geom
glm::mat4 invTranspose;
};
+struct bvhNode {
+ glm::vec3 aabbMin;
+ float _pad1;
+ glm::vec3 aabbMax;
+ float _pad2;
+ int startIdx;
+ int endIdx;
+ int leftChild;
+ int rightChild;
+ int isLeaf;
+};
+
struct Material
{
glm::vec3 color;
@@ -45,6 +64,11 @@ struct Material
float hasRefractive;
float indexOfRefraction;
float emittance;
+ struct
+ {
+ float iRad;
+ float oRad;
+ } blackHole;
};
struct Camera
@@ -83,5 +107,14 @@ struct ShadeableIntersection
{
float t;
glm::vec3 surfaceNormal;
+ int geoId;
int materialId;
};
+
+struct EnvironmentMap
+{
+ int width;
+ int height;
+ float* image;
+ int channels;
+};
diff --git a/src/tiny_obj_loader.h b/src/tiny_obj_loader.h
new file mode 100644
index 000000000..ee60f5383
--- /dev/null
+++ b/src/tiny_obj_loader.h
@@ -0,0 +1,3517 @@
+/*
+The MIT License (MIT)
+
+Copyright (c) 2012-Present, Syoyo Fujita and many contributors.
+
+Permission is hereby granted, free of charge, to any person obtaining a copy
+of this software and associated documentation files (the "Software"), to deal
+in the Software without restriction, including without limitation the rights
+to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+copies of the Software, and to permit persons to whom the Software is
+furnished to do so, subject to the following conditions:
+
+The above copyright notice and this permission notice shall be included in
+all copies or substantial portions of the Software.
+
+THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+THE SOFTWARE.
+*/
+
+//
+// version 2.0.0 : Add new object oriented API. 1.x API is still provided.
+// * Add python binding.
+// * Support line primitive.
+// * Support points primitive.
+// * Support multiple search path for .mtl(v1 API).
+// * Support vertex skinning weight `vw`(as an tinyobj
+// extension). Note that this differs vertex weight([w]
+// component in `v` line)
+// * Support escaped whitespece in mtllib
+// * Add robust triangulation using Mapbox
+// earcut(TINYOBJLOADER_USE_MAPBOX_EARCUT).
+// version 1.4.0 : Modifed ParseTextureNameAndOption API
+// version 1.3.1 : Make ParseTextureNameAndOption API public
+// version 1.3.0 : Separate warning and error message(breaking API of LoadObj)
+// version 1.2.3 : Added color space extension('-colorspace') to tex opts.
+// version 1.2.2 : Parse multiple group names.
+// version 1.2.1 : Added initial support for line('l') primitive(PR #178)
+// version 1.2.0 : Hardened implementation(#175)
+// version 1.1.1 : Support smoothing groups(#162)
+// version 1.1.0 : Support parsing vertex color(#144)
+// version 1.0.8 : Fix parsing `g` tag just after `usemtl`(#138)
+// version 1.0.7 : Support multiple tex options(#126)
+// version 1.0.6 : Add TINYOBJLOADER_USE_DOUBLE option(#124)
+// version 1.0.5 : Ignore `Tr` when `d` exists in MTL(#43)
+// version 1.0.4 : Support multiple filenames for 'mtllib'(#112)
+// version 1.0.3 : Support parsing texture options(#85)
+// version 1.0.2 : Improve parsing speed by about a factor of 2 for large
+// files(#105)
+// version 1.0.1 : Fixes a shape is lost if obj ends with a 'usemtl'(#104)
+// version 1.0.0 : Change data structure. Change license from BSD to MIT.
+//
+
+//
+// Use this in *one* .cc
+#define TINYOBJLOADER_IMPLEMENTATION
+// #include "tiny_obj_loader.h"
+//
+
+#ifndef TINY_OBJ_LOADER_H_
+#define TINY_OBJ_LOADER_H_
+
+#include